import numpy as np


class FilamentIndexer(object):

    def __init__(self, filament_id):
        self.filament_id = filament_id
        self._dict_time_index_to_indices = {}
        self.t_index_creation = None
        self.t_index_termination = None
        
    def set_indices_at_t_index(self, t_index: int, index_array: np.ndarray):
        self._dict_time_index_to_indices[t_index] = index_array
        if len(index_array) == 0:                        
            if self.t_index_creation is None:
                return
            elif self.t_index_termination is None:
                self.t_index_termination = t_index
            return
        if self.t_index_creation is None:
            self.t_index_creation = t_index

    def get_indices_at_t_index(self, t_index: int) -> np.ndarray:
        return self._dict_time_index_to_indices[t_index]

    def __getitem__(self, t_index: int) -> np.ndarray:
        return self._dict_time_index_to_indices[t_index]

    def __setitem__(self, t_index: int, index_array: np.ndarray):
        if not isinstance(t_index, int):
            object.__setattr__(self, t_index, index_array)
            return
        self.set_indices_at_t_index(t_index, index_array)

    def __str__(self):
        line = "{}: {},"
        s = "{"
        t_indices = list(self._dict_time_index_to_indices.keys())
        t_indices.sort()
        
        for t_index in t_indices[:3]:
            s += line.format(t_index, self._dict_time_index_to_indices[t_index])
            s += "\n"
        if len(t_indices) > 6:
            s += "...\n"
        for t_index in t_indices[-3:]:
            s += line.format(t_index, self._dict_time_index_to_indices[t_index])
            s += "\n"
        s = s[:-2]
        s += "}"
        return s

    def __repr__(self):
        return self.__str__()



from typing import Generator, Any, Tuple, List, Union, Dict
import numpy as np
import h5py
import ipdb
from .config_reader_beads import ConfigReaderBeads
from .text_stack_reader import (_FilamentsReader, _Beads, _Filaments)


class ChunkyTextStackConverter:

    def __init__(self, folder: str, chunk=50):
        self.folder = folder
        self.cfg_reader = ConfigReaderBeads(folder + '/config.json')
        self.h5_file = None
        self.freader = _FilamentsReader(folder + '/filaments.txt')
        self.freader.read()
        self.chunk_size = chunk
        self.breader = None

    def convert(self, fname_h5: str):
        self.h5_file = h5py.File(fname_h5)
        self.h5_file.attrs['FORMAT'] = "TEXT_STACK"

        config_content = open(self.folder + "/config.json", 'rt').read()
        self.h5_file.attrs['CONFIG'] = config_content

        filaments_array = self._convert_filaments_to_array()

        dset_filaments = self.h5_file.create_dataset("filaments", data=filaments_array,
                                                     compression='gzip', compression_opts=7)

        n_steps = filaments_array.shape[0]
        n_beads = filaments_array[:, :, 3].sum(1)
        n_beads_max = n_beads.max()

        print('n_steps =', n_steps)

        dset_bead_coords = self.h5_file.create_dataset("bead_coordinates",
                                                       shape=(n_steps, n_beads_max, 3),
                                                       compression='gzip',
                                                       compression_opts=7,
                                                       fillvalue=np.nan,
                                                       dtype='float64')

        dset_bead_links = self.h5_file.create_dataset("bead_links",
                                                      shape=(n_steps, n_beads_max, 3),
                                                      compression='gzip',
                                                      compression_opts=7,
                                                      fillvalue=-1,
                                                      dtype='int32')

        dset_bead_ids = self.h5_file.create_dataset('bead_ids',
                                                    shape=(n_steps, n_beads_max),
                                                    compression='gzip',
                                                    compression_opts=7,
                                                    fillvalue=-1,
                                                    dtype='int32')

        self.breader = _ChunkyBeadsReader(self.folder + '/beads.txt', self.chunk_size, n_beads_max)

        t_steps = []

        ti = 0
        for chunk_arrays in self.breader.read():
            ipdb.set_trace(context=5)
            t = chunk_arrays[0]
            ids = chunk_arrays[1]
            coords = chunk_arrays[2]
            links = chunk_arrays[3]

            t_steps.append(t)

            N = len(t)

            print("at chunk from {} to {}".format(ti, ti + N))

            dset_bead_ids[ti: ti+N] = ids
            dset_bead_coords[ti: ti+N] = coords
            dset_bead_links[ti: ti+N] = links
            ti = ti + N

        self.h5_file["steps"] = np.vstack(t_steps)
        self.h5_file.close()

    def _convert_filaments_to_array(self) -> np.ndarray:
        filaments_dict = self.freader.filaments
        N_max = self._determine_N_max_filaments(filaments_dict)
        filaments_array = np.full((len(filaments_dict.keys()), N_max, 4), -1, dtype=int)
        for i, ti in enumerate(sorted(filaments_dict.keys())):
            filaments = filaments_dict[ti]
            filaments_array[i, :len(filaments.ids), 0] = filaments.ids
            filaments_array[i, :len(filaments.ids), 1] = filaments.firsts
            filaments_array[i, :len(filaments.ids), 2] = filaments.lasts
            filaments_array[i, :len(filaments.ids), 3] = filaments.counts
        return filaments_array

    @staticmethod
    def _determine_N_max_filaments(filaments_dict: Dict[int, _Filaments]) -> int:
        N_max = 0
        for ti in filaments_dict:
            fils = filaments_dict[ti]
            if len(fils.ids) > N_max:
                N_max = len(fils.ids)
        return N_max

class _ChunkyBaseReader:

    def __init__(self, fname, chunksize=50):
        self.chunksize = chunksize
        self.fh = open(fname, 'rt')

    def read(self) -> Generator[List[np.ndarray], None, None]:
        while True:
            chunk_arrays = self._read_chunk()
            if not chunk_arrays:
                break
            yield chunk_arrays

    def _read_chunk(self) -> Union[List[np.ndarray], None]:
        count = 0
        t = []
        N = []
        data = []

        while count < self.chunksize:
            line = self.fh.readline()
            if line is None:
                break
            t_i, N_i = self.parse_step_line(line)
            t.append(t_i)
            N.append(N_i)
            data.append(self.parse_content_following_step_line([self.fh.readline() for i in range(N_i)]))
            count = count + 1
        if count == 0:
            return None
        return [np.array(t)] + self._data_list_to_chunk_array(data)

    def parse_content_following_step_line(self, content_lines) -> Any:
        raise NotImplementedError("Abstract Base Class")

    def _data_list_to_chunk_array(self, data: List[Any]) -> List[np.ndarray]:
        raise NotImplementedError("Abstract Base Class")

    @staticmethod
    def parse_step_line(step_line) -> Tuple[int, int]:
        split = step_line.split("\t")
        t = int(split[0].split("=")[1])
        N = int(split[1].split("=")[1])
        return t, N


class _ChunkyBeadsReader(_ChunkyBaseReader):

    def __init__(self, fname: str, chunksize: int, max_n_beads: int):
        _ChunkyBaseReader.__init__(self, fname, chunksize)
        self.max_n_beads = max_n_beads

    def parse_content_following_step_line(self, lines) -> _Beads:
        N = len(lines)
        ids = np.empty(N, dtype=int)
        x = np.empty(N)
        y = np.empty(N)
        z = np.empty(N)

        links_prev = np.empty(N, dtype=np.int16)
        links_next = np.empty(N, dtype=np.int16)
        links_crossf = np.empty(N, dtype=np.int16)
        for i, l in enumerate(lines):
            split = l.split("\t")
            ids[i] = int(split[0])
            x[i] = float(split[1])
            y[i] = float(split[2])
            z[i] = float(split[3])
            links_prev[i] = int(split[4])
            links_next[i] = int(split[5])
            links_crossf[i] = int(split[6])
        b = _Beads(ids, x, y, z, links_prev, links_next, links_crossf)

        return b

    def _data_list_to_chunk_array(self, beads_list: List[_Beads]) -> List[np.ndarray]:

        N = len(beads_list)
        coords_array = np.full((N, self.max_n_beads, 3), np.nan)
        link_array = np.full((N, self.max_n_beads, 3), -1, dtype=np.int32)
        bead_ids = np.full((N, self.max_n_beads), -1, dtype=int)

        for i, b in enumerate(beads_list):
            coords_array[i, :len(b.x), 0] = b.x
            coords_array[i, :len(b.x), 1] = b.y
            coords_array[i, :len(b.x), 2] = b.z
            link_array[i, :len(b.x), 0] = b.links_prev
            link_array[i, :len(b.x), 1] = b.links_next
            link_array[i, :len(b.x), 2] = b.links_crossf
            bead_ids[i, :len(b.x)] = b.ids
        return [bead_ids, coords_array, link_array]




from abc import ABC, abstractmethod
from typing import Tuple, Dict, List
import numpy as np

from actomyosin_analyser.model.bead import Filament


class DataReader(ABC):
    """
    Template of a data reader class. Any derived class should
    implement the here defined methods.
    This is a place holder to define the interface of a DataReader used
    by the Analyser class. Not all methods have to be overwritten, depending
    on the analyses you perform with the Analyser class.
    """

    @abstractmethod
    def __init__(self, source_file: str):
        ...

    @abstractmethod
    def read_particle_positions(self, minimum_image: bool) -> np.ndarray:
        ...

    @abstractmethod
    def read_time(self) -> np.ndarray:
        ...

    @abstractmethod
    def read_time_as_step_indices(self) -> np.ndarray:
        ...

    @abstractmethod
    def read_box_size(self) -> np.ndarray:
        """
        Read box size in shape (3, 2) where rows are x, y, z,
        columns are lower/upper bound.
        """
        ...

    @abstractmethod
    def read_particle_types(self) -> Tuple[Dict[str, int], np.ndarray]:
        """
        Read particle types as an integer array.
        :return: Dictionary mapping particle type (str) to type as integer.
        :return: Integer array of particle types.
        """
        ...

    @abstractmethod
    def get_filaments_all(self) -> List[List[Filament]]:
        """
        Read filament information for all frames.
        :return: Length of returned list is equal to number of recorded frames.
                 For each frame, each item is a list containing Filament instances.
        """
        ...

    @abstractmethod
    def read_dt(self):
        ...

    @abstractmethod
    def read_parameter(self, *args):
        ...




import numpy as np


def revert_minimum_image_projection(positions, box) -> np.ndarray:
    """
    :param positions: 3-Dimensional numpy array with axes (time, particle number, position)
    """
    box_indices = np.zeros_like(positions, dtype=int)
    displacements = positions[1:] - positions[:-1]
    box_index_changes = np.zeros_like(displacements, dtype=np.int8)
    box_index_changes[displacements > 0.5 * box] = -1
    box_index_changes[displacements < -0.5 * box] = 1
    box_indices[1:] = np.cumsum(box_index_changes, axis=0)
    reverted_positions = positions + box_indices * box

    return reverted_positions



from typing import Tuple
import h5py
import numpy as np
from ..model.bead_states import BeadStates


class HDF5BeadDataReader:

    def __init__(self, fname: str):
        self.h5_file = h5py.File(fname, "r")
        self.data_set_coords = self.h5_file.get("bead_coordinates")
        self.data_set_links = self.h5_file.get("bead_links")
        self.data_set_filaments = self.h5_file.get("filaments")
        self.data_set_steps = self.h5_file.get("steps")
        self.data_set_bead_ids = self.h5_file.get("bead_ids")
        
        if self.h5_file.attrs["FORMAT"] == "AFINES":
            self.afines_format = True
        else:
            self.afines_format = False

    def get_system_state_at_time_step(
            self, time_step: int) -> Tuple[np.ndarray, BeadStates]:

        coords_array = self.data_set_coords[time_step, :, :].flatten()
        links_array = self.data_set_links[time_step, :, :].flatten()
        bead_states = BeadStates(links_array)
        return coords_array, bead_states    



from typing import Tuple, List
import os

import numpy as np

from ._g_pipeline import _GPipeline
from .bead_msd import BeadMSD
from ...analysis.microrheology.mason_method import compute_G_advanced, compute_G_danny_seara
from ...tools.experiment_configuration import ExperimentConfiguration


class MasonMethod(_GPipeline):

    def __init__(self, experiment_configuration: ExperimentConfiguration,
                 skip: int,
                 method: str,
                 overwrite: bool=False):
        super().__init__(experiment_configuration, skip)
        self.plot_files.update({
            'G': os.path.join(experiment_configuration['GMason'], f'G_{method}.svg'),
            'G0': os.path.join(experiment_configuration['G0Mason'], f'G_with_G0_{method}.svg')
        })
        self.output_files.update({
            'G0': os.path.join(experiment_configuration['G0Mason'], f'G0_{method}.csv')
        })
        self._bead_msd = BeadMSD(experiment_configuration, skip, overwrite)
        self.skip = skip
        if method not in ['maier', 'seara']:
            raise ValueError("method can only be 'maier' or 'seara'")
        self._method = method
        if overwrite:
            self._remove_output_files()

    def run_analysis(self, **kwargs):
        omegas, Gs = self.get_omegas_and_Gs(**kwargs)
        self._plot_Gs(omegas, Gs, self.plot_files['G'])
        G0s = self._find_G0_via_saddle_point_method(omegas, Gs, self.output_files['G0'])
        self._plot_Gs_with_G0s(omegas, Gs, G0s, self.plot_files['G0'])

    def _validate_configuration(self):
        assert "GMason" in self.experiment_configuration
        assert "G0Mason" in self.experiment_configuration

    def get_omegas_and_Gs(self, **kwargs) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        if 'kT' not in kwargs:
            kwargs['kT'] = 1
        if 'dimensionality' not in kwargs:
            kwargs['dimensionality'] = 3
        msd_data = self._bead_msd.get_ensemble_msds()
        if self._method == 'maier':
            compute_function = compute_G_advanced
        elif self._method == 'seara':
            compute_function = compute_G_danny_seara
        else:
            raise RuntimeError("method has to be 'maier' or 'seara'")
        omegas, Gs = [], []
        groups = self.experiment_configuration.experiment_iterator
        for i, group in enumerate(groups):
            lag_time, msd = msd_data.get_lag_time_and_msd(i)
            r = msd_data.bead_radii[i]
            omega, G = compute_function(lag_time, msd, r, **kwargs)
            omegas.append(omega)
            Gs.append(G)
        return omegas, Gs



from math import isclose
import numpy as np
from mayavi.mlab import points3d, plot3d, gcf
from .mayavi_animator import MayaviAnimator
from ...model.bead import (STATE_FILAMENT_START,
                           STATE_FILAMENT_INTERIOR,
                           STATE_FILAMENT_END)


class BeadAnimator(MayaviAnimator):

    def __init__(self):
        self.data_reader = None
        self.beads = None
        self.tubes = []
        self.minimum_image_box = None
        self.bead_scale_factor = 20.0

    def show_at_t(self, t):
        if self.beads is not None:
            self.beads.scene.disable_render = True
        coordinates, status = self.data_reader.get_system_state_at_time_step(t)
        self._draw_data_slice(coordinates, status)
        if self.beads is not None:
            self.beads.scene.disable_render = False

    def _draw_data_slice(self, coordinates, bead_links):
        if coordinates.shape[0] % 3 != 0:
            msg = "coordinates' length needs to be multiple of 3 (x-, y-, z-coordinates)."
            raise RuntimeError(msg)
        x, y, z = self._data_slice_to_coords(coordinates)

        if self.minimum_image_box is not None:
            x -= self.minimum_image_box[0] * np.floor(x / self.minimum_image_box[0])
            y -= self.minimum_image_box[1] * np.floor(y / self.minimum_image_box[1])
            z -= self.minimum_image_box[2] * np.floor(z / self.minimum_image_box[2])
        
        s = bead_links.states
        if self.beads is None:
            size_array = self._bead_states_to_size_array(s)
            color_lookup_table = self._create_color_lookup_table()
            self.beads = points3d(x, y, z, size_array,
                                  scale_mode="scalar", scale_factor=self.bead_scale_factor,
                                  resolution=25)
            self.beads.module_manager.scalar_lut_manager.lut.table = color_lookup_table
            # ======================================================================================
            # connect filament beads with tubes
            # --------------------------------------------------------------------------------------
            for f in bead_links.filaments:                
                current = f.start
                indices = []
                while current != f.end:
                    indices.append(current)
                    current = bead_links.linked_beads[current].next
                indices.append(current)
                tube_coord_sets = self._get_minimum_image_tube_coords(
                    x[indices], y[indices], z[indices])
                for tube_set in tube_coord_sets:
                    x_tube = tube_set[0]
                    y_tube = tube_set[1]
                    z_tube = tube_set[2]
                    self.tubes.append(plot3d(x_tube, y_tube, z_tube,
                                             np.zeros_like(x_tube),
                                             color=(0.1725, 0.6274, 0.1725),
                                             tube_radius=2.0))
            # --------------------------------------------------------------------------------------
            return
        # ==========================================================================================
        # in case of animation, remove all tubes
        # ------------------------------------------------------------------------------------------
        for tubes in self.tubes:
            tubes.remove()
        self.tubes = []
        # ------------------------------------------------------------------------------------------
                
        self.beads.mlab_source.set(x=x, y=y, z=z)

    def initialize_animation(self, time_step):
        data_slice, status = self.data_reader.get_system_state_at_time_step(time_step)
        self._draw_data_slice(data_slice, status)
    
    def _get_minimum_image_tube_coords(self, x, y, z):
        """
        To connect beads at x, y, z with tubes, this method calculates
        where virtual beads need to be placed at the planes of periodic boundaries.
        """
        tube_coord_sets = []
        x_tube = [x[0]]
        y_tube = [y[0]]
        z_tube = [z[0]]
        n_virtual_beads = 0
        for i in range(1, x.shape[0]):
            dx = x[i] - x[i - 1]
            dy = y[i] - y[i - 1]
            dz = z[i] - z[i - 1]
            x_shift = self.minimum_image_box[0] * round(dx / self.minimum_image_box[0])
            y_shift = self.minimum_image_box[1] * round(dy / self.minimum_image_box[1])
            z_shift = self.minimum_image_box[2] * round(dz / self.minimum_image_box[2])
            boundary_cuts = {}          
            if x_shift != 0:            
                dx -= x_shift
                if (x[i - 1] + dx > self.minimum_image_box[0]):
                    r_cut_x = abs((self.minimum_image_box[0] - x[i - 1]) / dx)
                else:
                    r_cut_x = abs(x[i - 1] / dx)
                boundary_cuts['x'] = r_cut_x
            if y_shift != 0:
                dy -= y_shift
                if (y[i - 1] + dy > self.minimum_image_box[1]):
                    r_cut_y = abs((self.minimum_image_box[1] - y[i - 1]) / dy)
                else:
                    r_cut_y = abs(y[i - 1] / dy)
                boundary_cuts['y'] = r_cut_y
            if z_shift != 0:
                dz -= z_shift
                if (z[i - 1] + dz > self.minimum_image_box[2]):
                    r_cut_z = abs((self.minimum_image_box[2] - z[i - 1]) / dz)
                else:
                    r_cut_z = abs(z[i - 1] / dz)
                boundary_cuts['z'] = r_cut_z
            sorted_boundary_cut_pairs = sorted(boundary_cuts.items(), key=lambda x: x[1])
            prev_x = x[i - 1]
            prev_y = y[i - 1]
            prev_z = z[i - 1]
            prev_r = 0
            for pair in sorted_boundary_cut_pairs:                
                key, r = pair                
                virtual_x = prev_x + (r - prev_r) * dx
                virtual_y = prev_y + (r - prev_r) * dy
                virtual_z = prev_z + (r - prev_r) * dz
                n_virtual_beads += 1
                x_tube.append(virtual_x)
                y_tube.append(virtual_y)
                z_tube.append(virtual_z)                
                tube_coord_sets.append((np.array(x_tube), np.array(y_tube), np.array(z_tube)))

                if key == 'x':
                    if isclose(virtual_x, 0.0, abs_tol=self.minimum_image_box[0] * 0.01):
                        virtual_x += self.minimum_image_box[0]
                    else:
                        virtual_x -= self.minimum_image_box[0]
                elif key == 'y':
                    if isclose(virtual_y, 0.0, abs_tol=self.minimum_image_box[1] * 0.01):
                        virtual_y += self.minimum_image_box[1]
                    else:
                        virtual_y -= self.minimum_image_box[1]
                elif key == 'z':
                    if isclose(virtual_z, 0.0, abs_tol=self.minimum_image_box[2] * 0.01):
                        virtual_z += self.minimum_image_box[2]
                    else:
                        virtual_z -= self.minimum_image_box[2]
                x_tube = [virtual_x]
                y_tube = [virtual_y]
                z_tube = [virtual_z]
                prev_x = virtual_x
                prev_y = virtual_y
                prev_z = virtual_z
                prev_r = r
            x_tube.append(x[i])
            y_tube.append(y[i])
            z_tube.append(z[i])
        tube_coord_sets.append((np.array(x_tube), np.array(y_tube), np.array(z_tube)))

        return tube_coord_sets

    def _bead_states_to_size_array(self, states):
        size_array = np.zeros_like(states)
        size_array[states == STATE_FILAMENT_START] = 1.25
        size_array[states == STATE_FILAMENT_END] = 1.25
        size_array[states == STATE_FILAMENT_INTERIOR] = 1.0
        return size_array

    def _create_color_lookup_table(self):
        lut = np.zeros((255, 4), dtype=np.int8)
        lut[254, :] = [31, 119, 180, 255]
        lut[0, :] = [0, 0, 0, 255]
        lut[1: 254, 0] = np.ones_like(lut[1: 254, 0]) * 255
        lut[1: 254, 1] = np.ones_like(lut[1: 254, 1]) * 127
        lut[1: 254, 2] = np.ones_like(lut[1: 254, 2]) * 14
        lut[1: 254, 3] = np.ones_like(lut[1: 254, 3]) * 255
        return lut



import os
from configparser import ConfigParser
import numpy as np
import h5py

from typing import Tuple, Dict

from afines_reader.actins_reader import ActinsReader
from afines_reader.motors_reader import MotorsReader
from afines_reader.model.actins import Actins
from afines_reader.model.motors import Motors

class AfinesConverter:

    def __init__(self, afines_output_dir):
        self.afines_dir = os.path.abspath(afines_output_dir)

        self.actins_reader = ActinsReader(self.afines_dir + "/txt_stack/actins.txt")
        self.pmotors_reader = MotorsReader(self.afines_dir + "/txt_stack/pmotors.txt")
        self.amotors_reader = MotorsReader(self.afines_dir + "/txt_stack/amotors.txt")

        self.actins_reader.read()
        self.pmotors_reader.read()
        self.amotors_reader.read()

        self.h5_file = None
        self.config = ConfigParser()

    def convert(self, fname_h5: str):        
        self.h5_file = h5py.File(fname_h5)

        self.h5_file.attrs['FORMAT'] = "AFINES"
        config_content = open(self.afines_dir + "/data/config_full.cfg", 'rt').read()
        self.h5_file.attrs['CONFIG'] = config_content
        # insert dummy section header, otherwise can't use ConfigParser
        self.config.read_string("[DUMMY]\n" + config_content)
        
        bead_array, link_array, filament_array = self._convert_actins_to_arrays()
        self.h5_file["bead_coordinates"] = bead_array
        self.h5_file["bead_links"] = link_array
        self.h5_file["filaments"] = filament_array

        bead_ids = self._create_full_bead_ids_array(bead_array.shape[0], bead_array.shape[1])
        self.h5_file["bead_ids"] = bead_ids

        xlink_coords, xlink_links = self._convert_motors_to_arrays(self.pmotors_reader.motors)
        self.h5_file["crosslink_coordinates"] = xlink_coords
        self.h5_file["crosslink_links"] = xlink_links

        motor_coords, motor_links = self._convert_motors_to_arrays(self.amotors_reader.motors)
        self.h5_file["motor_coordinates"] = motor_coords
        self.h5_file["motor_links"] = motor_links

        time_steps = self._convert_t_array_to_steps()
        self.h5_file["steps"] = time_steps
                
        self.h5_file.close()

    def _create_full_bead_ids_array(self, n_time_steps, n_beads):
        return np.mgrid[:n_time_steps, :n_beads][1].astype(np.int32)
        
    def _convert_t_array_to_steps(self) -> np.ndarray:
        t_actins  = np.array(sorted(self.actins_reader.actins.keys()))
        t_amotors = np.array(sorted(self.amotors_reader.motors.keys()))
        t_pmotors = np.array(sorted(self.pmotors_reader.motors.keys()))

        try:
            assert (t_actins  == t_amotors).all()
            assert (t_actins  == t_pmotors).all()
            assert (t_amotors == t_pmotors).all()
        except AssertionError as error:
            msg = "Time points read from actins.txt, amotors.txt and pmotors.txt do not all match!"
            raise RuntimeError(msg)

        dt = float(self.config["DUMMY"]["dt"])
        
        t = t_actins
        t_as_steps = np.around(t / dt).astype(np.int)
        return t_as_steps        
        
    def _convert_motors_to_arrays(
            self, motors_dict: Dict[float, Motors]) -> Tuple[np.ndarray, np.ndarray]:
        actins_dict = self.actins_reader.actins
        N_max = self._determine_N_max_motors(motors_dict)        
        coords_array = np.full((len(motors_dict.keys()), N_max, 4), np.nan)
        links_array  = np.full((len(motors_dict.keys()), N_max, 2), -1, dtype=np.int32)
        for i, ti in enumerate(sorted(motors_dict.keys())):
            motors = motors_dict[ti]
            actins = actins_dict[ti]
            coords_array[i, :len(motors.x), 0] = motors.x
            coords_array[i, :len(motors.x), 1] = motors.y
            coords_array[i, :len(motors.x), 2] = motors.dx
            coords_array[i, :len(motors.x), 3] = motors.dy
            links_array[i, :len(motors.x), 0] = self._get_bead_index_from_fidx_and_lidx(
                motors.fidx0, motors.lidx0, actins)
            links_array[i, :len(motors.x), 1] = self._get_bead_index_from_fidx_and_lidx(
                motors.fidx1, motors.lidx1, actins)
        return coords_array, links_array

    def _get_bead_index_from_fidx_and_lidx(
            self, fidx: np.ndarray, lidx: np.ndarray, actins: Actins) -> np.ndarray:
        bead_idx = np.empty_like(fidx, dtype=np.int32)
        for i in range(len(fidx)):        
            fidx_i = fidx[i]
            if fidx_i == -1 or lidx[i] == -1:
                bead_idx[i] = -1
                continue
            beads_with_fidx = np.where(actins.fidx == fidx_i)[0]
            bead_idx[i] = beads_with_fidx[lidx[i]]
        return bead_idx

    def _determine_N_max_motors(self, motors_dict: Dict[float, Motors]) -> int:
        N_max = 0
        for ti in motors_dict:
            motors = motors_dict[ti]
            N = len(motors.x)
            if N > N_max:
                N_max = N
        return N_max

    def _convert_actins_to_arrays(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:        
        actins_dict = self.actins_reader.actins
        N_max_beads, N_max_filaments = self._determine_N_max_beads_and_filaments(actins_dict)
        bead_array = np.full((len(actins_dict.keys()), N_max_beads, 2), np.nan)
        link_array = np.full((len(actins_dict.keys()), N_max_beads, 2), -1, dtype=np.int32)
        filament_array = np.full((len(actins_dict.keys()), N_max_filaments, 3), -1, dtype=np.int32)
        for i, ti in enumerate(sorted(actins_dict.keys())):
            actins = actins_dict[ti]
            if (actins.r != actins.r[0]).any():
                msg = "Radii of beads are not uniform! Need to store that information somewhere."
                raise NotImplementedError(msg)
            bead_array[i, :len(actins.x), 0] = actins.x
            bead_array[i, :len(actins.x), 1] = actins.y
            unique_fidx = np.unique(actins.fidx)
            for j, fidx in enumerate(unique_fidx):
                beads_with_fidx = np.where(actins.fidx == fidx)[0]
                if len(beads_with_fidx) < 2:
                    continue
                assert (beads_with_fidx == np.arange(beads_with_fidx[0], beads_with_fidx[-1] + 1))\
                    .all()
                filament_array[i, j, 0] = fidx
                filament_array[i, j, 1] = beads_with_fidx[0]
                filament_array[i, j, 2] = beads_with_fidx[-1]
                link_array[i, beads_with_fidx[0]: beads_with_fidx[-1], 1] =\
                    beads_with_fidx[0: -1] + 1
                link_array[i, beads_with_fidx[1]: beads_with_fidx[-1] + 1, 0] =\
                    beads_with_fidx[0: -1]
        return bead_array, link_array, filament_array
                
    def _determine_N_max_beads_and_filaments(
            self, actins_dict: Dict[float, Actins]) -> Tuple[int, int]:
        N_max_beads = 0
        N_max_filaments = 0
        for ti in actins_dict:
            actins = actins_dict[ti]
            if len(actins.x) > N_max_beads:
                N_max_beads = len(actins.x)
            N_filaments = len(np.unique(actins.fidx))
            if N_filaments > N_max_filaments:
                N_max_filaments = N_filaments
        return N_max_beads, N_max_filaments



import vtk

color = vtk.vtkNamedColors()

def draw_outline_with_lines(x_min, x_max, y_min, y_max, z_min, z_max):
    corner_points = vtk.vtkPoints()
    corner_points.InsertNextPoint([0., 0., 0.]) # 0
    corner_points.InsertNextPoint([0., 0., z_max]) # 1
    corner_points.InsertNextPoint([0., y_max, 0.]) # 2
    corner_points.InsertNextPoint([x_max, 0., 0.]) # 3
    corner_points.InsertNextPoint([x_max, 0., z_max]) # 4
    corner_points.InsertNextPoint([x_max, y_max, 0]) # 5
    corner_points.InsertNextPoint([0., y_max, z_max]) # 6
    corner_points.InsertNextPoint([x_max, y_max, z_max]) # 7

    line0 = vtk.vtkLine()
    line0.GetPointIds().SetId(0, 0)
    line0.GetPointIds().SetId(1, 1)

    line1 = vtk.vtkLine()
    line1.GetPointIds().SetId(0, 0)
    line1.GetPointIds().SetId(1, 2)

    line2 = vtk.vtkLine()
    line2.GetPointIds().SetId(0, 0)
    line2.GetPointIds().SetId(1, 3)

    line3 = vtk.vtkLine()
    line3.GetPointIds().SetId(0, 1)
    line3.GetPointIds().SetId(1, 4)

    line4 = vtk.vtkLine()
    line4.GetPointIds().SetId(0, 1)
    line4.GetPointIds().SetId(1, 6)

    line5 = vtk.vtkLine()
    line5.GetPointIds().SetId(0, 2)
    line5.GetPointIds().SetId(1, 5)
    
    line6 = vtk.vtkLine()
    line6.GetPointIds().SetId(0, 2)
    line6.GetPointIds().SetId(1, 6)

    line7 = vtk.vtkLine()
    line7.GetPointIds().SetId(0, 3)
    line7.GetPointIds().SetId(1, 4)

    line8 = vtk.vtkLine()
    line8.GetPointIds().SetId(0, 3)
    line8.GetPointIds().SetId(1, 5)

    line9 = vtk.vtkLine()
    line9.GetPointIds().SetId(0, 4)
    line9.GetPointIds().SetId(1, 7)

    line10 = vtk.vtkLine()
    line10.GetPointIds().SetId(0, 5)
    line10.GetPointIds().SetId(1, 7)

    line11 = vtk.vtkLine()
    line11.GetPointIds().SetId(0, 6)
    line11.GetPointIds().SetId(1, 7)

    lines = vtk.vtkCellArray()
    lines.InsertNextCell(line0)
    lines.InsertNextCell(line1)
    lines.InsertNextCell(line2)
    lines.InsertNextCell(line3)
    lines.InsertNextCell(line4)
    lines.InsertNextCell(line5)
    lines.InsertNextCell(line6)
    lines.InsertNextCell(line7)
    lines.InsertNextCell(line8)
    lines.InsertNextCell(line9)
    lines.InsertNextCell(line10)
    lines.InsertNextCell(line11)

    lines_poly_data = vtk.vtkPolyData()
    lines_poly_data.SetPoints(corner_points)
    lines_poly_data.SetLines(lines)

    colors = vtk.vtkUnsignedCharArray()
    colors.SetNumberOfComponents(3)
    colors.SetName("Colors")

    for i in range(12):
        colors.InsertNextTypedTuple([0, 240, 0])

    lines_poly_data.GetCellData().SetScalars(colors)

    mapper = vtk.vtkPolyDataMapper()
    mapper.SetInputData(lines_poly_data)

    actor = vtk.vtkActor()
    actor.SetMapper(mapper)

    return actor


def draw_outline_with_tubes(x_min: float, x_max: float,
                            y_min: float, y_max: float,
                            z_min: float, z_max: float,
                            r: float=10.0) -> vtk.vtkActor:    
    
    cube = vtk.vtkCubeSource()
    cube.SetBounds(x_min, x_max, y_min, y_max, z_min, z_max)

    edges = vtk.vtkExtractEdges()
    edges.SetInputConnection(cube.GetOutputPort())

    tubes = vtk.vtkTubeFilter()
    tubes.SetInputConnection(edges.GetOutputPort())
    tubes.SetRadius(r)
    tubes.SetNumberOfSides(6)
    tubes.UseDefaultNormalOn()
    tubes.SetDefaultNormal(.577, .577, .577)

    tube_mapper = vtk.vtkPolyDataMapper()
    tube_mapper.SetInputConnection(tubes.GetOutputPort())

    actor = vtk.vtkActor()
    actor.SetMapper(tube_mapper)
    actor.GetProperty().SetDiffuseColor(color.GetColor3d("khaki"))

    return actor

def draw_plane() -> vtk.vtkActor:
    plane_source = vtk.vtkPlaneSource()
    plane_source.SetNormal(0, 0, 1)
    plane_source.SetCenter(550, 550, 0)
    actor = _poly_data_to_actor(plane_source)
    actor.GetProperty().SetColor(color.GetColor3d("gray"))
    return actor
    

def _poly_data_to_actor(poly_data) -> vtk.vtkActor:
    mapper = vtk.vtkPolyDataMapper()
    mapper.SetInputConnection(poly_data.GetOutputPort())

    actor = vtk.vtkActor()
    actor.SetMapper(mapper)
    return actor



import vtk


class VTKFilamentRepresentation:

    def __init__(self):
        self.sphere_mapper = []
        self.sphere_actors = []
        self.sphere_sources = []
        self.vtk_renderer = None

    def set_coordinates(self, data_slice):
        if data_slice.shape[0] % 3 != 0:
            msg = "data_slice needs to be multiple of 3 (x-, y-, z-coordinates)."
            raise RuntimeError(msg)
        for i in range(0, data_slice.shape[0] // 3):            
            coords = tuple(data_slice[i * 3: i * 3 + 3])
            if len(self.sphere_actors) <= i:
                new_sphere = vtk.vtkSphereSource()
                new_sphere.SetCenter(*coords)
                new_sphere.SetRadius(20)

                mapper = vtk.vtkPolyDataMapper()
                mapper.SetInputConnection(new_sphere.GetOutputPort())
                new_actor = vtk.vtkActor()
                new_actor.SetMapper(mapper)
                self.sphere_actors.append(new_actor)
                self.sphere_mapper.append(mapper)
                self.sphere_sources.append(new_sphere)
                self.vtk_renderer.AddActor(new_actor)
            else:                
                self.sphere_sources[i].SetCenter(*coords)



from typing import List, Union
import os

import numpy as np
import pandas as pd

from .displacement_distribution import DisplacementDistribution, _GaussFitResult
from .filament_trajectories import FilamentTrajectories
from .._pipeline import Pipeline
from ..time import Time
from ...tools.experiment_configuration import ExperimentConfiguration
from ...tools.experiment_iterator import GroupIterator

_BooleanArray = np.ndarray


class ExplosionDetector(Pipeline):

    def run_analysis(self, fixed_filament_length: int, n_bins_histogram: int):
        displacement_pipeline = DisplacementDistribution(self.experiment_configuration)
        gauss_fits = displacement_pipeline.get_gauss_fits(fixed_filament_length, n_bins_histogram)
        passed_gauss_fit_check = self._perform_gauss_fit_check(gauss_fits)
        passed_filament_displacement_check = self._perform_filament_displacement_check(gauss_fits,
                                                                                       passed_gauss_fit_check,
                                                                                       fixed_filament_length)

        self._write_results_table(passed_gauss_fit_check, passed_filament_displacement_check)

    def get_number_of_displacements_exceeding_threshold(self, group, sigma, fixed_filament_length) -> np.ndarray:
        threshold = sigma * self.sigma_threshold_single_filaments
        exceeding = self._compute_number_of_displacements_exceeding_threshold(group, threshold, fixed_filament_length)
        return exceeding

    def _validate_configuration(self):
        assert "ExplosionCheck" in self.experiment_configuration

    def __init__(self, experiment_configuration: ExperimentConfiguration,
                 sigma_threshold_gauss_fit: float=5.,
                 sigma_threshold_single_filaments: float=5.,
                 threshold_n_exceeding_per_filament: int=2,
                 threshold_n_filaments_exceeding_per_simulation: int=2,
                 skip_n_frames: int=5):
        super().__init__(experiment_configuration)
        self.sigma_threshold_gauss_fit = sigma_threshold_gauss_fit
        self.sigma_threshold_single_filaments = sigma_threshold_single_filaments
        self.threshold_n_exceeding_per_filament = threshold_n_exceeding_per_filament
        self.threshold_n_filaments_exceeding_per_simulation = threshold_n_filaments_exceeding_per_simulation
        self.skip_n_frames = skip_n_frames

        self.output_files.update({
            "ExplosionCheck": os.path.join(self.experiment_configuration["ExplosionCheck"], 'explosion_check.csv')
        })
        self.trajectories_pipeline = FilamentTrajectories(experiment_configuration)

    def _perform_gauss_fit_check(self, gauss_fits: List[_GaussFitResult]) -> List[bool]:
        results = []
        for i, group in enumerate(self.experiment_configuration.experiment_iterator.groups):
            fit_result = gauss_fits[i]
            if fit_result is None:
                results.append(False)
                continue
            sigma = fit_result.optimized_parameters[1]
            r = self._gauss_fit_check_single_group(sigma, group)
            results.append(r)

        return results

    def _gauss_fit_check_single_group(self, sigma: float, group: GroupIterator) -> bool:
        t = Time(self.experiment_configuration).get_time(group)
        diff_t = t[1:] - t[:-1]
        if not np.isclose(diff_t, diff_t[0]).all():
            raise RuntimeError("Can't handle dynamic time deltas / frame rates.")
        delta_t = diff_t[0]
        expected_sigma = np.sqrt(6 * delta_t)
        if sigma > self.sigma_threshold_gauss_fit*expected_sigma:
            return False
        return True

    def _write_results_table(self, passed_gauss_fit_check: List[bool],
                             passed_filament_displacement_check: List[bool]):
        passed_gauss_fit_check = np.array(passed_gauss_fit_check)
        passed_filament_displacement_check = np.array(passed_filament_displacement_check)
        table = pd.DataFrame(index=self.experiment_configuration.experiment_iterator.create_index_from_groups(),
                             columns=['all checks passed', 'Gauss-fit-check passed',
                                      'filament-displacement-check passed'])
        table['Gauss-fit-check passed'] = passed_gauss_fit_check
        table['filament-displacement-check passed'] = passed_filament_displacement_check
        table['all checks passed'] = (passed_filament_displacement_check & passed_gauss_fit_check)

        table.to_csv(self.output_files["ExplosionCheck"])

    def _perform_filament_displacement_check(self, gauss_fits: List[_GaussFitResult],
                                             passed_previous_checks: List[bool],
                                             fixed_filament_length: int) -> List[Union[bool, None]]:
        groups = self.experiment_configuration.experiment_iterator.groups
        results = []
        for i, group in enumerate(groups):
            if not passed_previous_checks[i]:
                results.append(None)
                continue
            gf = gauss_fits[i]
            sigma = gf.optimized_parameters[1]
            r = self._filament_displacement_check_single_group(group, sigma, fixed_filament_length)
            results.append(r)
        return results

    def _filament_displacement_check_single_group(self, group: GroupIterator, sigma: float,
                                                  fixed_filament_length: int) -> bool:
        n_exceeding = self.get_number_of_displacements_exceeding_threshold(group, sigma, fixed_filament_length)
        n_exceeding_too_large = n_exceeding > self.threshold_n_exceeding_per_filament
        n_filaments_exceeding = np.sum(n_exceeding_too_large, axis=1)
        if (n_filaments_exceeding > self.threshold_n_filaments_exceeding_per_simulation).any():
            return False
        return True

    def _compute_number_of_displacements_exceeding_threshold(
            self, group, threshold, fixed_filament_length) -> np.ndarray:
        times, trajectories = self.trajectories_pipeline.get_trajectories_of_single_group(
            group, fixed_filament_length, [0, fixed_filament_length - 1])

        n_particles = trajectories.shape[2]

        displacements = np.abs(trajectories[:, self.skip_n_frames + 1:] - trajectories[:, self.skip_n_frames: -1])
        boolean = displacements > threshold
        n_exceeded = np.sum(boolean, axis=(1, 3))

        heads = n_exceeded[:, np.arange(0, n_particles, 2)]
        tails = n_exceeded[:, np.arange(1, n_particles, 2)]
        total = heads + tails

        return total
        # sim_fil_indices_exceeding = np.zeros((n_simulations, n_filaments), dtype=bool)
        # return sim_fil_indices_exceeding




import os
from ..file_io.data_reader import DataReader
from ..tools.experiment_iterator import ExperimentIterator


class ExperimentConfiguration:

    def __init__(
            self,
            root_folder: str,
            result_folder: str,
            data_reader_class: DataReader,
            experiment_iterator: ExperimentIterator,
            skip_n_frames: int
    ):
        self.root_folder = root_folder
        self.result_folder = result_folder
        self.data_reader_class = data_reader_class
        self.experiment_iterator = experiment_iterator
        self.skip_n_frames = skip_n_frames

        self._sub_folders = {}

    def add_result_sub_folder(self, key: str, sub_path: str, create: bool=True):
        if key in self._sub_folders:
            raise KeyError(f"Sub folder for {key} is already defined: {self._sub_folders[key]}")
        self._sub_folders[key] = os.path.join(self.result_folder, sub_path)
        if not create:
            return
        os.makedirs(self._sub_folders[key], exist_ok=True)

    def get_result_sub_folder(self, key) -> str:
        return self._sub_folders[key]

    def __getitem__(self, key) -> str:
        return self.get_result_sub_folder(key)

    def __setitem__(self, key, value):
        self.add_result_sub_folder(key, value, create=True)

    def __contains__(self, key):
        return key in self._sub_folders



import numpy as np
import glob
import os
from ..model.bead_states import BeadStates

class TextBeadDataReader:

    def __init__(self, folder):
        self.coord_files = glob.glob(folder + "/bead_coords_*.txt")
        self.coord_files.sort()        
        if len(self.coord_files) == 0:
            raise RuntimeError("No actin coordinates in folder '" + folder + "'")
        self.link_files = glob.glob(folder + "/bead_links_*.txt")
        self.link_files.sort()
        self.lines_in_files = np.ones(len(self.coord_files), dtype=np.int64)
        self.lines_in_files = self.lines_in_files * -1

        for i in range(len(self.coord_files)):
            coord_file = self.coord_files[i]
            fname, ext = os.path.splitext(coord_file)
            if os.path.isfile(fname + ".npy"):
                self.coord_files[i] = fname + ".npy"

        for j in range(len(self.link_files)):
            link_file = self.link_files[j]
            fname, ext = os.path.splitext(link_file)
            if os.path.isfile(fname + ".npy"):
                self.link_files[j] = fname + ".npy"                
        
        self._open_files(0)

    def _open_files(self, i):
        if i >= len(self.coord_files):
            raise RuntimeError("index " + str(i) + " exceeds number of files")
        for j in range(i):
            if (self.lines_in_files[j] == -1):
                message = "Trying to open file " + str(i) + ", but file "\
                          + str(j) + " has never been opened.\n"
                message += "Files need to be opened sequentially to get"
                message += "the mapping of the time_step right."
                raise RuntimeError(message)
        if TextBeadDataReader._is_npy_file(self.coord_files[i]):
            self.array_current_file = np.load(self.coord_files[i])
        else:
            self.array_current_file = np.genfromtxt(self.coord_files[i])
        if TextBeadDataReader._is_npy_file(self.link_files[i]):
            self.array_current_links = np.load(self.link_files[i])
        else:
            self.array_current_links = np.genfromtxt(self.link_files[i], dtype=int)        
        self.current_file = i
        self.lines_in_files[i] = self.array_current_file.shape[0]
        self.first_time_step_of_current_array = np.sum(self.lines_in_files[:i])

    def get_system_state_at_time_step(self, time_step):
        file_index = self._locate_file_containing_time_step(time_step)
        if (file_index == -1):
            raise RuntimeError("time_step " + str(time_step) + " is out of range")
        if self.current_file != file_index:
            self._open_files(file_index)
        coords_array = self.array_current_file[time_step - self.first_time_step_of_current_array]
        links_array = self.array_current_links[time_step - self.first_time_step_of_current_array]
        bead_links = BeadStates(links_array)
        return coords_array, bead_links

    def _locate_file_containing_time_step(self, time_step):
        _sum_lines = 0
        for i in range(len(self.lines_in_files)):
            if (self.lines_in_files[i] == -1):
                self._open_files(i)
            _sum_lines = np.sum(self.lines_in_files[:i + 1])
            if (_sum_lines > time_step):
                return i
        return -1

    def convert_text_files_to_npy(self):
        for txt_file in self.coord_files:
            fname, ext = os.path.splitext(txt_file)
            if ext != ".txt":
                continue
            a = np.loadtxt(txt_file)
            np.save(fname + ".npy", a)

        for txt_file in self.link_files:
            fname, ext = os.path.splitext(txt_file)
            if ext != ".txt":
                continue
            a = np.loadtxt(txt_file, dtype=int)
            np.save(fname + ".npy", a)

    def _is_npy_file(f):
        fname, ext = os.path.splitext(f)
        if ext == ".npy":
            return True
        return False



from typing import List
import os
import numpy as np
import pandas as pd

BooleanArray = np.ndarray


def merge_and_save_table(filename: str, table: pd.DataFrame):
    index_cols = list(range(table.index.nlevels))
    if not os.path.exists(filename):
        table.to_csv(filename)
        return
    old_table = pd.read_csv(filename, index_col=index_cols)
    merged_table = _merge_tables(table, old_table)
    merged_table.to_csv(filename)


def _merge_tables(new_table: pd.DataFrame, old_table: pd.DataFrame) -> pd.DataFrame:
    assert (new_table.columns == old_table.columns).all()
    merged_index = _merge_indices(new_table.index, old_table.index)
    df = pd.DataFrame(index=merged_index, columns=new_table.columns)
    df.loc[new_table.index] = new_table
    valid_rows = _get_rows_with_data(old_table)
    diff_index = old_table.index[valid_rows].difference(new_table.index)
    df.loc[diff_index] = old_table.loc[diff_index]
    return df


def _merge_indices(index1, index2) -> pd.Index:
    for name in index1.names:
        assert name in index2.names
        assert index1.get_level_values(name).dtype == index2.get_level_values(name).dtype
    assert len(index1.names) == len(index2.names)
    merged = index1.append(index2.difference(index1))
    return merged


def _get_rows_with_data(table: pd.DataFrame) -> BooleanArray:
    a = np.array(~table.isna())
    return a.sum(axis=1) == table.shape[1]



from typing import List
import os
import numpy as np
import matplotlib.pyplot as plt
from .._pipeline import Pipeline
from ...tools.experiment_configuration import ExperimentConfiguration
from ...tools.experiment_iterator import GroupIterator


class BeadDisplacement(Pipeline):

    def __init__(self, experiment_configuration: ExperimentConfiguration):
        super().__init__(experiment_configuration)
        self.plot_files.update({
            'BeadDisplacement': os.path.join(experiment_configuration['BeadDisplacement'],
                                             'bead_displacement_{label}.svg')
        })

    def _validate_configuration(self):
        assert "BeadDisplacement" in self.experiment_configuration

    def run_analysis(self, *args, **kwargs):
        displacements = self._get_displacements()
        self._plot_displacements(displacements)

    def _get_displacements(self) -> List[List[np.ndarray]]:
        displacements = []
        for group in self.experiment_configuration.experiment_iterator:
            displacements_group = BeadDisplacement._get_displacements_group(group)
            displacements.append(displacements_group)
        return displacements

    @staticmethod
    def _get_displacements_group(group: GroupIterator) -> List[np.ndarray]:
        displacements = []
        for sim in group:
            a = sim.analyser
            pos = a.get_trajectories_non_filament()[:, 0]
            displacements_sim = np.sqrt(np.sum((pos[1:] - pos[:-1])**2, axis=1))
            displacements.append(displacements_sim)
        return displacements

    def _plot_displacements(self, displacements: List[List[np.ndarray]]):
        groups = self.experiment_configuration.experiment_iterator.groups
        for i, g in enumerate(groups):
            fig, ax = plt.subplots(1, 1)
            title = g.label
            file_label = g.get_label_from_values()
            d_group = displacements[i]
            for d, idx in zip(d_group, g.index):
                ax.plot(d, label=f'sim {idx:03}')
            ax.set(
                title=title,
                xlabel='frame',
                ylabel='displacements / $x_0$'
            )
            fig.tight_layout()
            filename = self.plot_files['BeadDisplacement'].format(label=file_label)
            fig.savefig(filename)




import numpy as np
import h5py
from typing import Tuple, Dict
from .text_stack_reader import TextStackReader, _Beads, _Filaments
from .config_reader_beads import ConfigReaderBeads


class TextStackConverter:

    def __init__(self, folder: str):
        self.folder = folder
        self.ts_reader = TextStackReader(folder)
        self.cfg_reader = ConfigReaderBeads(folder + "/config.json")
        self.h5_file = None

    def convert(self, fname_h5: str):
        self.h5_file = h5py.File(fname_h5)
        self.h5_file.attrs['FORMAT'] = "TEXT_STACK"

        config_content = open(self.folder + "/config.json", 'rt').read()
        self.h5_file.attrs['CONFIG'] = config_content

        bead_ids, bead_array, link_array = self._convert_beads_to_arrays()
        filaments_array = self._convert_filaments_to_array()

        dset_bead_coords = self.h5_file.create_dataset("bead_coordinates", data=bead_array,
                                                       compression='gzip', compression_opts=7)
        dset_bead_links = self.h5_file.create_dataset("bead_links", data=link_array,
                                                      compression='gzip', compression_opts=7)
        dset_filaments = self.h5_file.create_dataset("filaments", data=filaments_array,
                                                     compression='gzip', compression_opts=7)
        dset_bead_ids = self.h5_file.create_dataset("bead_ids", data=bead_ids,
                                                    compression='gzip', compression_opts=7)

        self.h5_file["steps"] =\
            np.array(sorted(self.ts_reader.beads_reader.beads.keys()))

        self.h5_file.close()
        
    def _convert_beads_to_arrays(self) \
        -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        beads_dict = self.ts_reader.beads_reader.beads
        N_max = self._determine_N_max_beads(beads_dict)
        coords_array = np.full((len(beads_dict.keys()), N_max, 3), np.nan)
        link_array = np.full((len(beads_dict.keys()), N_max, 3), -1, dtype=np.int32)
        bead_ids = np.full((len(beads_dict.keys()), N_max), -1, dtype=int)
        for i, ti in enumerate(sorted(beads_dict.keys())):
            beads = beads_dict[ti]
            coords_array[i, :len(beads.x), 0] = beads.x
            coords_array[i, :len(beads.x), 1] = beads.y
            coords_array[i, :len(beads.x), 2] = beads.z
            link_array[i, :len(beads.x), 0] = beads.links_prev
            link_array[i, :len(beads.x), 1] = beads.links_next
            link_array[i, :len(beads.x), 2] = beads.links_crossf
            bead_ids[i, :len(beads.x)] = beads.ids
        return bead_ids, coords_array, link_array
            

    def _convert_filaments_to_array(self) -> np.ndarray:
        filaments_dict = self.ts_reader.filaments_reader.filaments
        N_max = self._determine_N_max_filaments(filaments_dict)
        filaments_array = np.full((len(filaments_dict.keys()), N_max, 4), -1, dtype=int)
        for i, ti in enumerate(sorted(filaments_dict.keys())):
            filaments = filaments_dict[ti]
            filaments_array[i, :len(filaments.ids), 0] = filaments.ids
            filaments_array[i, :len(filaments.ids), 1] = filaments.firsts
            filaments_array[i, :len(filaments.ids), 2] = filaments.lasts
            filaments_array[i, :len(filaments.ids), 3] = filaments.counts
        return filaments_array

    @staticmethod
    def _determine_N_max_beads(beads_dict: Dict[int, _Beads]) -> int:
        N_max = 0
        for ti in beads_dict:
            beads = beads_dict[ti]
            if len(beads.x) > N_max:
                N_max = len(beads.x)
        return N_max

    @staticmethod
    def _determine_N_max_filaments(filaments_dict: Dict[int, _Filaments]) -> int:
        N_max = 0
        for ti in filaments_dict:
            fils = filaments_dict[ti]
            if len(fils.ids) > N_max:
                N_max = len(fils.ids)
        return N_max
        



import numpy as np
from math import isclose
from typing import List


def split_filament_coords_at_borders(ordered_bead_coords: np.ndarray, box) -> List[np.ndarray]:
    """
    @param ordered_bead_coords Coordinates of beads, have to be corrected via minimum image!
    """
    coordinate_sets = []
    diff = ordered_bead_coords[1:] - ordered_bead_coords[:-1]
    shifts = _compute_shifts_array(diff, box)
    shifts_at = np.where(shifts != 0)
    if len(shifts_at[0]) == 0:
        return [ordered_bead_coords]
    shifts_at_n, shifts_at_i, r_cuts = _sort_conflicting_shifts(shifts_at[0], shifts_at[1],
                                                                ordered_bead_coords, diff,
                                                                shifts, box)
    shifted_diffs = diff + shifts
    n_prev = shifts_at_n[0]
    prev_virtual_bead = None

    coord_set, prev_virtual_bead = _generate_coord_set_before_first_split(
        ordered_bead_coords[:shifts_at_n[0] + 1], n_prev, r_cuts[0], shifted_diffs)
    coordinate_sets.append(coord_set)
    for j, n in enumerate(shifts_at_n[1:]):
        j = j+1
        set_shape = list(ordered_bead_coords.shape)
        if n == n_prev:
            num_points = 2
        else:
            num_points = n - n_prev + 2
        set_shape[0] = num_points
        coord_set = np.empty(tuple(set_shape))
        if prev_virtual_bead is not None:
            prev_i = shifts_at_i[j-1]
            if isclose(prev_virtual_bead[prev_i], box[prev_i], abs_tol=box[prev_i] * 0.01):
                prev_virtual_bead[prev_i] -= box[prev_i]
            else:
                prev_virtual_bead[prev_i] += box[prev_i]
            coord_set[0] = prev_virtual_bead
        if n == n_prev:
            virtual_bead = coord_set[0] + (r_cuts[j] - r_cuts[j-1]) * shifted_diffs[n]
        else:
            coord_set[1: -1] = ordered_bead_coords[n_prev+1: n+1]
            virtual_bead = ordered_bead_coords[n] + r_cuts[j] * shifted_diffs[n]
        prev_virtual_bead = virtual_bead
        coord_set[-1] = virtual_bead

        coordinate_sets.append(coord_set)
        n_prev = n
        
    coord_set = _generate_coord_set_after_last_split(
        ordered_bead_coords, n_prev, shifts_at_i, prev_virtual_bead, box)
    
    coordinate_sets.append(coord_set)
    
    return coordinate_sets

def _generate_coord_set_before_first_split(bead_coords_before_split, n, r_cut, shifted_diffs):
    coord_set_shape = list(bead_coords_before_split.shape)
    coord_set_shape[0] = coord_set_shape[0] + 1
    coord_set = np.empty(tuple(coord_set_shape))
    coord_set[:-1] = bead_coords_before_split[:]
    virtual_bead = bead_coords_before_split[-1] + r_cut * shifted_diffs[n]
    coord_set[-1] = virtual_bead
    return coord_set, virtual_bead
    

def _generate_coord_set_after_last_split(ordered_bead_coords, n_prev,
                                         shifts_at_i, prev_virtual_bead, box):
    num_points = len(ordered_bead_coords) - n_prev
    set_shape = list(ordered_bead_coords.shape)
    set_shape[0] = num_points
    coord_set = np.empty(tuple(set_shape))
    last_i = shifts_at_i[-1]
    if isclose(prev_virtual_bead[last_i], box[last_i], abs_tol=box[last_i] * 0.01):
        prev_virtual_bead[last_i] -= box[last_i]
    else:
        prev_virtual_bead[last_i] += box[last_i]
    coord_set[0] = prev_virtual_bead
    coord_set[1:] = ordered_bead_coords[n_prev+1:]
    return coord_set

def _compute_shifts_array(diff, box):
    shifts = np.zeros_like(diff)
    for i, b in enumerate(box):
        if b is None:            
            continue
        shifts_i = -b * np.rint(diff[:, i] / b)
        shifts[:, i] = shifts_i
    return shifts
        
def _sort_conflicting_shifts(shifts_at_n, shifts_at_i,
                             ordered_bead_coords, diff, shifts, box):

    r_cuts = np.zeros(shifts_at_n.shape)
    for j in range(len(shifts_at_n)):
        conflicting_indices = []
        nj = shifts_at_n[j]
        k = j + 1
        while k < len(shifts_at_n):
            nk = shifts_at_n[k]
            if nj != nk:
                break
            conflicting_indices.append(k)
            k += 1
        if len(conflicting_indices) == 0:
            ij = shifts_at_i[j]
            dx = diff[nj, ij] + shifts[nj, ij]
            x_nj_ij = ordered_bead_coords[nj, ij]
            if x_nj_ij + dx > box[ij]:
                r_cut = abs((box[ij] - x_nj_ij) / dx)
            else:                
                r_cut = abs(x_nj_ij / dx)
            r_cuts[j] = r_cut
            continue
        conflicting_indices.insert(0, j)
        boundary_cuts = {}
        for l in conflicting_indices:
            nl = shifts_at_n[l]
            il = shifts_at_i[l]
            dx = diff[nl, il] + shifts[nl, il]
            x_nl_il = ordered_bead_coords[nl, il]
            if x_nl_il + dx > box[il]:
                r_cut = abs((box[il] - x_nl_il) / dx)
            else:                
                r_cut = abs(x_nl_il / dx)
            boundary_cuts[l] = r_cut
        sorted_boundary_cut_pairs = sorted(boundary_cuts.items(), key=lambda x: x[1])

        conflicting_shifts_in_correct_order_n = []
        conflicting_shifts_in_correct_order_i = []
        r_cuts_in_correct_order = []
        # a cut pair is a Tuple of the index in the shifts_at array and the r_cut value
        for cut_pair in sorted_boundary_cut_pairs:
            conflicting_shifts_in_correct_order_n.append(shifts_at_n[cut_pair[0]])
            conflicting_shifts_in_correct_order_i.append(shifts_at_i[cut_pair[0]])
            r_cuts_in_correct_order.append(cut_pair[1])
        shifts_at_n[j:j+len(conflicting_indices)] = conflicting_shifts_in_correct_order_n
        shifts_at_i[j:j+len(conflicting_indices)] = conflicting_shifts_in_correct_order_i
        r_cuts[j:j+len(conflicting_indices)] = r_cuts_in_correct_order
    return shifts_at_n, shifts_at_i, r_cuts



from typing import Union
import numpy as np
from scipy.integrate import simps

a = 14.054
b = 0.473
cij = np.array([[-0.75, 0.359375, -0.109375],
                [-0.5, 1.0625, -0.5625]])

def _c(k: float) -> float:
    return 1 - (1 + (0.38*k**-0.95)**-5)**-0.2

def _d(k: float) -> float:
    if k < 0.125:
        return 0.0
    return 1-1/(0.177/(k - 0.111) + 6.4 * (k-0.111)**0.783)

def _sum1(i: int, k: float, r: float) -> float:
    return cij[i+1, 0] * k**i * r ** 2 + cij[i+1, 1] * k**i * r ** 4 + cij[i+1, 2] * k**i * r ** 6

def _J_SYD(k: float) -> float:
    if k <= 0.125:
        return (3/4/np.pi/k)**1.5 * np.exp(0) * (1-1.25*k)
    else:
        return 112.04 * k**2 * np.exp(0.246/k - a*k)

def _QI(r: Union[float, np.ndarray], k: float) -> Union[float, np.ndarray]:
    return _J_SYD(k) * ((1-_c(k)*r**2)/(1-r**2))**2.5 * np.exp((_sum1(-1, k, r)+_sum1(0, k, r))/(1-r**2))\
        * np.exp(-(_d(k)*k * a * b * (1+b) * r**2)/(1-b**2*r**2)) * np.i0(-(_d(k)*k*a*(1+b)*r)/(1-b**2*r**2))

def CDF_QI(r: np.ndarray, k: float) -> np.ndarray:
    dQI = []
    for i in range(1, len(r)):
        dQIi = simps(_QI(r[:i], k) * r[:i] ** 2, r[:i])
        dQI.append(dQIi)
    CDF = np.array(dQI)
    return CDF/CDF[-1]

def mean_relative_end_to_end_distance(k: float, num_points: int=500) -> float:
    r = np.linspace(0, 1, num_points)
    qi = _QI(r, k)
    if np.isnan(qi[-1]):
        qi = qi[:-1]
        r = r[:-1]
    R = 4*np.pi*simps(qi * r**3, r)
    return R


import vtk


class BeadAnimator:

    def __init__(self):
        self.data_reader = None
        self.bead_sources = []
        self.vtk_renderer = None
        self.bead_radius = None
        self.minimum_image = False

    def show_at_t(self, t):
        coordinates, status = self.data_reader\
                                  .get_system_state_at_time_step(t, self.minimum_image)
        self._draw_data_slice(coordinates, status)

    def _draw_data_slice(self, coordinates, status):
        if coordinates.shape[0] % 3 != 0:
            msg = "data_slice needs to be multiple of 3 (x-, y-, z-coordinates)."
            raise RuntimeError(msg)

        if len(self.bead_sources) == 0:
            for i in range(0, coordinates.shape[0] // 3):            
                coords = tuple(coordinates[i * 3: i * 3 + 3])
                new_sphere = vtk.vtkSphereSource()
                new_sphere.SetCenter(*coords)
                new_sphere.SetRadius(self.bead_radius)

                mapper = vtk.vtkPolyDataMapper()
                mapper.SetInputConnection(new_sphere.GetOutputPort())
                new_actor = vtk.vtkActor()
                new_actor.SetMapper(mapper)
                self.bead_sources.append(new_sphere)
                self.vtk_renderer.AddActor(new_actor)
            return
        for i in range(0, coordinates.shape[0] // 3):
            coords = tuple(coordinates[i * 3: i * 3 + 3])
            self.bead_sources[i].SetCenter(*coords)



import numpy as np
from mayavi.mlab import points3d, plot3d
from .mayavi_animator import MayaviAnimator


class SegmentAnimator(MayaviAnimator):

    def __init__(self):
        self.data_reader = None
        self.points_begin = None
        self.points_end = None
        self.tubes = []

    def show_at_t(self, t, mlab_fig=None):
        self._set_render_disable_status(True)
        data_slice_t = self.data_reader.get_coordinates_at_time_step(t)
        self._draw_data_slice(data_slice_t)
        self._set_render_disable_status(False)

    def _set_render_disable_status(self, state: bool):
        self.points_begin.scene.disable_render = state
        self.points_end.scene.disable_render = state
        for tube in self.tubes:
            tube.scene.disable_render = state

    def initialize_animation(self, time_step):
        data_slice = self.data_reader.get_coordinates_at_time_step(time_step)
        self._draw_data_slice(data_slice)

    def _draw_data_slice(self, data_slice):        
        if data_slice.shape[0] % 3 != 0:
            msg = "data_slice needs to be multiple of 3 (x-, y-, z-coordinates)."
            raise RuntimeError(msg)
        if data_slice.shape[0] % 2 != 0:
            msg = "data_slice needs to have even number of values (2 points per segment)"
            raise RuntimeError(msg)
        x, y, z = self._data_slice_to_coords(data_slice)

        ind_even = np.arange(0, x.shape[0], 2)
        ind_odd = np.arange(1, x.shape[0], 2)

        if self.points_begin is None:
            self.points_begin = points3d(x[ind_even], y[ind_even], z[ind_even], color=(1.0, 0., 0.),
                                         scale_mode="none", scale_factor=20.0, resolution=25)
        else:
            self.points_begin.mlab_source.set(x=x[ind_even], y=y[ind_even], z=z[ind_even])
        if self.points_end is None:
            self.points_end = points3d(x[ind_odd], y[ind_odd], z[ind_odd], color=(0.0, 1., 0.),
                                       scale_mode="none", scale_factor=20.0, resolution=25)
        else:
            self.points_end.mlab_source.set(x=x[ind_odd], y=y[ind_odd], z=z[ind_odd])

        if len(self.tubes) == 0:
            for i in range(0, x.shape[0], 2):
                current_x = np.array([x[i], x[i + 1]])
                current_y = np.array([y[i], y[i + 1]])
                current_z = np.array([z[i], z[i + 1]])
                self.tubes.append(plot3d(current_x, current_y, current_z,
                                         tube_radius=10., tube_sides=15))
        else:
            for i in range(0, x.shape[0], 2):
                tube = self.tubes[i // 2]
                current_x = np.array([x[i], x[i + 1]])
                current_y = np.array([y[i], y[i + 1]])
                current_z = np.array([z[i], z[i + 1]])
                tube.mlab_source.set(x=current_x, y=current_y, z=current_z)



from typing import Callable, Dict, Any, Iterable, List, Tuple
from warnings import warn
import math
import json
import numpy as np
import numba
import h5py
from actomyosin_analyser.file_io.data_reader import DataReader
from . import geometry
from ..__init__ import _VERSION
from ..model.bead import Filament

Dataset = h5py.Dataset
_FilamentT = Tuple[np.ndarray, np.ndarray]
_FilamentsT = numba.typed.List  # [_FilamentT]


class Analyser:

    def __init__(self, data_reader: DataReader, analysis_file: str):
        self.data_reader = data_reader
        self.analysis_file = h5py.File(analysis_file, 'a')
        self._offset_filaments = None
        self._particle_positions = None

    def _get_analysis_dataset(self,
                              dset_name: str,
                              compute_function: Callable[[], np.ndarray],
                              **keyword_arguments: Dict[str, Any]) -> np.ndarray:
        kwargs_as_json = None
        if keyword_arguments:
            kwargs_as_json = Analyser._keyword_arguments_to_json_string(keyword_arguments)
            dset_name += '_' + kwargs_as_json
        if dset_name in self.analysis_file:
            dset = self.analysis_file[dset_name]
            self._check_version(dset)
            if keyword_arguments:                
                kwargs = json.loads(dset.attrs['parameters'])
                if kwargs != keyword_arguments:
                    msg = "DataSet '{}' has matching name, but loaded parameters are different!"
                    msg = msg.format(dset.name)
                    raise RuntimeError(msg)
            return np.array(dset)
        data = compute_function(**keyword_arguments)
        self.analysis_file[dset_name] = data
        self.analysis_file[dset_name]\
            .attrs['ACTOMYOSIN_ANALYSER_VERSION'] = _VERSION
        if kwargs_as_json is not None:            
            self.analysis_file[dset_name].attrs['parameters'] = kwargs_as_json
        return data

    def _get_nested_analysis_dataset(self, group_name: str,
                                     compute_function: Callable[[], np.ndarray],
                                     **keyword_arguments: Dict[str, Any]) -> np.ndarray:
        if not keyword_arguments:
            msg = "keyword_arguments can not be empty for Analyser._get_nested_analysis_dataset. "
            msg += "If you have no keywords, use Analyser._get_analysis_dataset instead!"
            raise KeyError(msg)
        kwargs_as_json = Analyser._keyword_arguments_to_json_string(keyword_arguments)
        # hasher = hashlib.sha1(kwargs_as_json)
        # h = hasher.hexdigest()
        h = kwargs_as_json

        if group_name in self.analysis_file:
            group = self.analysis_file[group_name]
        else:
            group = self.analysis_file.create_group(group_name)

        if h in group:
            dset = group[h]
            self._check_version(dset)
            kwargs = json.loads(dset.attrs['parameters'])
            for_comparison = keyword_arguments.copy()
            for k in for_comparison.keys():
                value = for_comparison[k]
                if isinstance(value, tuple):
                    for_comparison[k] = 'hashed_tuple_' + str(hash(value))
            if kwargs != for_comparison:
                msg = "DataSet '{}' has matching name, but loaded parameters are different!".format(
                    dset.name)
                raise RuntimeError(msg)
            return np.array(dset)

        data = compute_function(**keyword_arguments)
        group[h] = data
        group[h].attrs['parameters'] = kwargs_as_json
        group[h].attrs['ACTOMYOSIN_ANALYSER_VERSION'] = _VERSION
        return data

    def _get_parameter(self, parameter_name,
                       compute_function: Callable[[], Any]) -> Any:
        attrs = self.analysis_file.attrs
        if parameter_name in attrs:
            return attrs[parameter_name]
        par = compute_function()
        attrs[parameter_name] = par
        return par

    def get_offset_filaments(self) -> int:
        return self._get_parameter('offset_filaments',
                                   self._determine_offset_filaments)
        
    def _determine_offset_filaments(self) -> int:
        if self._offset_filaments is not None:
            return self._offset_filaments
        self._offset_filaments = self.data_reader.get_n_non_filament_particles()
        return self._offset_filaments

    def get_filaments(self, save: bool=False) -> List[List[Filament]]:
        return self._get_filaments_from_analysis_file_or_data_reader(save)

    def get_trajectories_filaments(self, minimum_image: bool=False) -> np.ndarray:
        return self._get_analysis_dataset('trajectories_filaments',
                                          self._get_trajectories_filament,
                                          minimum_image=minimum_image)

    def get_trajectories_non_filament(self, minimum_image=False) -> np.ndarray:
        return self._get_analysis_dataset('trajectories_non_filament',
                                          self._get_trajectories_non_filament,
                                          minimum_image=minimum_image)

    def get_msd_ensemble(self, particle_indices: List[int], skip=0) -> np.ndarray:
        return self._get_nested_analysis_dataset('msd_ensemble', self._compute_msd_ensemble,
                                                 particle_indices=tuple(particle_indices),
                                                 skip=skip)


    def get_msd(self, particle_indices: List[int], non_filament: bool=False, skip=0) -> np.ndarray:
        if non_filament:
            method = self.get_msd_single_particle_non_filament
        else:
            method = self.get_msd_single_particle
        msd0 = method(particle_indices[0], skip)
        msds = np.full((len(particle_indices), msd0.shape[0], msd0.shape[1]), np.nan)
        msds[0] = msd0
        for i in range(1, len(particle_indices)):
            msds[i] = method(particle_indices[i], skip)
        return msds
    
    def get_msd_single_particle(self, particle_index: int, skip: int) -> np.ndarray:
        return self._get_nested_analysis_dataset('msd_filament', self._compute_msd,
                                                 particle_index=particle_index,
                                                 skip=skip)

    def get_msd_single_particle_non_filament(self, particle_index: int, skip: int) -> np.ndarray:
        return self._get_nested_analysis_dataset('msd_non_filament', self._compute_msd_non_filament,
                                                 particle_index=particle_index,
                                                 skip=skip)

    def get_densities(self, particle_types: Iterable[str], n_bins: np.ndarray,
                      ranges: np.ndarray=None, count: bool=False) -> np.ndarray:
        n_bins, range_x, range_y, range_z =\
            self._convert_binning_parameters_to_lists(n_bins, ranges)
        return self._get_nested_analysis_dataset('densities', self._compute_densities,
                                                 particle_types=particle_types,
                                                 range_x=range_x,
                                                 range_y=range_y,
                                                 range_z=range_z,
                                                 n_bins=n_bins,
                                                 count=count)

    def get_densities_gauss_kernel(self, particle_types: List[str],
                                   n_bins: np.ndarray,
                                   ranges: np.ndarray,
                                   sigmas: np.ndarray) -> np.ndarray:
        n_bins, range_x, range_y, range_z =\
            self._convert_binning_parameters_to_lists(n_bins, ranges)
        sigmas = list(sigmas)
        return self._get_nested_analysis_dataset(
            'densities_gauss_kernel',
            self._compute_densities_gauss_kernel,
            particle_types=particle_types,
            range_x=range_x,
            range_y=range_y,
            range_z=range_z,
            n_bins=n_bins,
            sigmas=sigmas
        )

    def get_link_count(self) -> np.ndarray:
        return self._get_analysis_dataset('link_count', self.data_reader.get_motor_count)

    def get_contour_lengths(self) -> np.ndarray:
        return self._get_analysis_dataset('contour_lengths',
                                          self._compute_contour_lenghts)
    
    def get_end_to_end_distances(self, t_frame: int, n_beads: int) -> np.ndarray:
        """
        Get end-to-end distances of filaments with n_beads at frame t_frame.
        """
        return self._get_nested_analysis_dataset('end_to_end_distances',
                                                 self._compute_end_to_end_distances,
                                                 t=t_frame,
                                                 n_beads=n_beads)

    def get_end_to_end_vectors(self) -> np.ndarray:
        """
        Get end-to-end vectors for all filaments at all frames. 
        
        :return: Array with 3 dimensions: (1) time frame (2) filament ID (3) XYZ coordinates.
        """
        return self._get_analysis_dataset('end_to_end_vectors', self._get_end_to_end_vectors)

    def get_time_in_steps(self) -> np.ndarray:
        return self._get_analysis_dataset('time_in_steps', self._get_time_in_steps)

    def get_energies_stretch_and_bend(self) -> np.ndarray:
        """
        Get bending and stretching energy of filaments.
        """
        energies = self._get_analysis_dataset('energies', self._compute_energies_stretch_and_bend)
        return energies

    def _get_time_in_steps(self):
        return self.data_reader.read_time_as_step_indices()
        
    def _get_trajectories_non_filament(self, minimum_image):
        pos = self.data_reader.read_particle_positions(minimum_image)
        offset = self.get_offset_filaments()
        return pos[:, :offset]

    def _get_trajectories_filament(self, minimum_image: bool):
        pos = self.data_reader.read_particle_positions(minimum_image)
        offset = self.get_offset_filaments()
        return pos[:, offset:]
        
    def _convert_binning_parameters_to_lists(self, n_bins, ranges):
        if ranges is None:
            ranges = self.data_reader.read_box_size()
        if isinstance(ranges, np.ndarray):
            range_x = ranges[0].tolist()
            range_y = ranges[1].tolist()
            range_z = ranges[2].tolist()
        else:
            range_x = list(ranges[0])
            range_y = list(ranges[1])
            range_z = list(ranges[2])
        if isinstance(n_bins, np.ndarray):
            n_bins = n_bins.tolist()
        n_bins = list(n_bins)
        return n_bins, range_x, range_y, range_z

    def _compute_msd(self, particle_index: int, skip: int) -> np.ndarray:
        coords = self.get_trajectories_filaments()[skip:, particle_index]
        return Analyser._compute_msd_from_trajectory(coords)

    def _compute_msd_non_filament(self, particle_index: int, skip: int) -> np.ndarray:
        offset = self.get_offset_filaments()
        if particle_index >= offset:
            msg = "Filaments start at particle index {}.".format(offset)
            msg += " Non-filament particles can't have index larger than that."
            msg += " Your requested particle index: {}".format(particle_index)
            raise KeyError(msg)
        traj = self.get_trajectories_non_filament()[skip:, particle_index]
        return Analyser._compute_msd_from_trajectory(traj)

    def _compute_msd_ensemble(self, particle_indices: Tuple[int], skip: int):
        coords = self.get_trajectories_filaments()[skip:, particle_indices]
        return Analyser._compute_msd_from_trajectory(coords)

    @staticmethod
    @numba.njit
    def _compute_msd_from_trajectory(trajectory: np.ndarray) -> np.ndarray:
        lags = np.arange(1, len(trajectory))
        msd = np.empty((len(lags), 3))
        for lag in lags:
            displacements = trajectory[lag:] - trajectory[:-lag]
            squared_summed = np.sum(displacements**2, axis=-1)
            msd[lag-1, 0] = lag
            msd[lag-1, 1] = np.nanmean(squared_summed)
            msd[lag-1, 2] = np.nanvar(squared_summed)
        return msd

    def _compute_densities(self,
                           particle_types: List[str],
                           range_x: List[float],
                           range_y: List[float],
                           range_z: List[float],
                           n_bins: List[int],
                           count: bool) -> np.ndarray:
        wx = (range_x[1] - range_x[0]) / n_bins[0]
        wy = (range_y[1] - range_y[0]) / n_bins[1]
        wz = (range_z[1] - range_z[0]) / n_bins[2]
        V = wx*wy*wz

        positions = self.data_reader.read_particle_positions()
        pmap, ptypes = self.data_reader.read_particle_types()

        shape = (len(positions), n_bins[0], n_bins[1], n_bins[2])
        densities = np.full(shape, np.nan)

        for t in range(positions.shape[0]):
            pos_t = positions[t]
            ptypes_t = ptypes[t]
            mask = np.zeros_like(ptypes_t, dtype=bool)
            for pt in particle_types:
                mask += (ptypes_t == pmap[pt])
            selected_particles = pos_t[mask]
            bin_indices = Analyser._coordinates_to_bin_indices(selected_particles,
                                                               range_x, wx,
                                                               range_y, wy,
                                                               range_z, wz)
            for i in range(shape[1]):
                for j in range(shape[2]):
                    for k in range(shape[3]):   
                        count_ijk = ((bin_indices[:, 0] == i) &
                                 (bin_indices[:, 1] == j) &
                                 (bin_indices[:, 2] == k)).sum()
                        densities[t, i, j, k] = count_ijk
        if count:
            densities = densities.astype(int)
        else:
            densities = densities/V
        
        return densities

    def _compute_densities_gauss_kernel(self,
                                        particle_types: List[str],
                                        range_x: List[float],
                                        range_y: List[float],
                                        range_z: List[float],
                                        n_bins: List[int],
                                        sigmas: List[float]) -> np.ndarray:
        wx = (range_x[1] - range_x[0]) / n_bins[0]
        wy = (range_y[1] - range_y[0]) / n_bins[1]
        wz = (range_z[1] - range_z[0]) / n_bins[2]
        V = wx*wy*wz


        box = self.data_reader.read_box_size()
        box_l = np.array([
            box[1, 0] - box[0,0],
            box[1, 1] - box[0, 1],
            box[1, 2] - box[0, 2]
        ])

        grid_range = (np.array(sigmas)*4 / np.array([wx, wy, wz])).astype(int)

        positions = self.data_reader.read_particle_positions()
        pmap, ptypes = self.data_reader.read_particle_types()

        shape = (len(positions), n_bins[0], n_bins[1], n_bins[2])
        densities = np.full(shape, np.nan)

        for t in range(positions.shape[0]):
            pos_t = positions[t]
            ptypes_t = ptypes[t]
            mask = np.zeros_like(ptypes_t, dtype=bool)
            for pt in particle_types:
                mask += (ptypes_t == pmap[pt])
            selected_particles = pos_t[mask]
            bin_indices = Analyser._coordinates_to_bin_indices(selected_particles,
                                                               range_x, wx,
                                                               range_y, wy,
                                                               range_z, wz)

            for i in range(shape[1]):
                for j in range(shape[2]):
                    for k in range(shape[3]):
                        mask_ijk = self._get_voxel_mask(
                            bin_indices, grid_range, (i, j, k), shape[1:]
                        )
                        pos_voxel = np.array([
                            (i+0.5) * wx + range_x[0],
                            (j+0.5) * wy + range_y[0],
                            (k+0.5) * wz + range_z[0]
                        ])
                        delta_x = geometry.get_minimum_image_vector(
                            selected_particles[mask_ijk] + box_l/2,
                            pos_voxel+box_l/2,
                            box_l
                        )
                        kernel_density_ijk = Analyser._gauss_3D(delta_x, sigmas).sum()
                        densities[t, i, j, k] = kernel_density_ijk
        return densities

    def _compute_end_to_end_distances(self, t: int, n_beads: int) -> np.ndarray:
        coordinates = self.get_trajectories_filaments()[t]
        filaments = self.data_reader.get_filaments(t)
        box = self.data_reader.read_box_size()
        box = np.array([box[1, 0] - box[0, 0],
                        box[1, 1] - box[0, 1],
                        box[1, 2] - box[0, 2]])
    
        e2e = []
        for fil in filaments:
            if len(fil.items) != n_beads:
                continue
            e2e.append(self._compute_e2e_single_filament(coordinates, fil, box))
        return np.array(e2e)

    def _get_end_to_end_vectors(self) -> np.ndarray:
        raise NotImplementedError
        coordinates = self.get_trajectories_filaments()
        filaments = self.data_reader.get_filaments_all()


    def _compute_e2e_single_filament(self, coordinates: np.ndarray, filament, box) -> float:
        offset_filaments = self.get_offset_filaments()
        fcoords = coordinates[np.array(filament.items)+offset_filaments]
        v_segments = geometry.get_minimum_image_vector(fcoords[:-1], fcoords[1:], box)
        e2e = np.sum(v_segments, axis=0)
        return math.sqrt(e2e[0]**2 + e2e[1]**2 + e2e[2]**2)

    def _compute_contour_lenghts(self) -> np.ndarray:
        coordinates = self.get_trajectories_filaments()
        box = self.data_reader.read_box_size()
        box = np.array([box[1, 0] - box[0, 0],
                        box[1, 1] - box[0, 1],
                        box[1, 2] - box[0, 2]])

        filaments = self.get_filaments()
        
        contour_lengths = np.full((len(coordinates), len(filaments[0])), np.nan)

        for t in range(len(coordinates)):
            fil_t = filaments[t]
            for f, fil in enumerate(fil_t):
                fcoords = coordinates[t, fil.items]
                v_segments = geometry.get_minimum_image_vector(fcoords[:-1], fcoords[1:], box)
                d = np.sqrt(np.sum(v_segments**2, 1))
                lc = np.sum(d)
                contour_lengths[t, f] = lc

        return contour_lengths

    def _compute_energies_stretch_and_bend(self) -> np.ndarray:
        coordinates = self.get_trajectories_filaments()
        box = self.data_reader.read_box_size()
        box = np.array([box[1, 0] - box[0, 0],
                        box[1, 1] - box[0, 1],
                        box[1, 2] - box[0, 2]])

        filaments = self.get_filaments()
        filaments_as_tuples = Analyser._convert_nested_filaments_to_tuples(filaments)

        return Analyser._numba_compute_energies_stretch_and_bend(box, coordinates,
                                                                 filaments_as_tuples)

    @staticmethod
    def _convert_nested_filaments_to_tuples(filaments) -> numba.typed.List:
        filaments_as_tuples = numba.typed.List()
        for f_frame in filaments:
            filaments_as_tuples_single_frame = numba.typed.List()
            for f in f_frame:
                filaments_as_tuples_single_frame.append((
                    np.array(f.items).astype('uint32'),
                    np.array(f.motors).astype('uint32')))
            filaments_as_tuples.append(filaments_as_tuples_single_frame)
        return filaments_as_tuples

    @staticmethod
    @numba.njit
    def _numba_compute_energies_stretch_and_bend(
            box, coordinates,
            filaments: numba.typed.List  # List[_FilamentsT]
    ):
        energies = np.zeros((len(coordinates), 2))
        for t in range(len(coordinates)):
            fil_t = filaments[t]
            energies_t = np.zeros((len(fil_t), 2))
            for f, fil in enumerate(fil_t):
                coords_t = coordinates[t]
                fcoords = coords_t[fil[0]]
                v_segments = geometry.get_minimum_image_vector(
                    fcoords[:-1], fcoords[1:], box)
                segment_lengths = np.sqrt(np.sum(v_segments ** 2, 1))

                e_stretch = np.sum((segment_lengths - 1) ** 2)

                cos_theta = ((v_segments[1:] * v_segments[:-1]).sum(1)
                             / segment_lengths[1:] / segment_lengths[:-1])

                e_bend = np.sum(np.arccos(cos_theta) ** 2)

                energies_t[f, 0] = e_stretch
                energies_t[f, 1] = e_bend

            energies[t] = np.sum(energies_t, axis=0)
        return energies

    @staticmethod
    def _keyword_arguments_to_json_string(keyword_arguments: Dict[str, Any]) -> str:
        """
        Note that tuples get hashed. This is to avoid containers with thousands of values
        to become part of the json string, which will be used as key to retrieve
        data from file. Make sure you provide containers with countless items as tuples.
        """
        if not keyword_arguments:
            return ''
        sorted_dict = {}
        for key in sorted(keyword_arguments.keys()):
            value = keyword_arguments[key]
            if isinstance(value, tuple):
                value = 'hashed_tuple_' + str(hash(value))
            sorted_dict[key] = value
        return json.dumps(sorted_dict)

    @staticmethod
    def _check_version(data_set: Dataset):
        if "ACTOMYOSIN_ANALYSER_VERSION" in data_set.attrs:
            if data_set.attrs['ACTOMYOSIN_ANALYSER_VERSION'] != _VERSION:
                msg = "!!!! DataSet '" + data_set.name + "' was created with "
                msg += "actomyosin_analyser version " \
                       + data_set.attrs['ACTOMYOSIN_ANALYSER_VERSION']
                msg += "!\nYour version is " + _VERSION + ".\n"
                msg += "Make sure that the data in here is what you expect it to be."
                warn(msg, RuntimeWarning)
            return
        msg = "!!!! DataSet '" + data_set.name + "' has no version attribute.\n"
        msg += "Make sure that the data in here is what you expect it to be."
        warn(msg, RuntimeWarning)

    @staticmethod
    def _gauss_3D(delta_x: np.ndarray, sigma: np.ndarray) -> np.ndarray:
        return np.exp(-((delta_x/sigma)**2/2).sum(1))

    @staticmethod
    def _get_voxel_mask(bin_indices, grid_range, indices, shape) -> np.ndarray:
        i, j, k = indices
        mask_i = ((np.abs((bin_indices[:, 0]) - i) <= grid_range[0]) |
                  ((-np.abs(bin_indices[:, 0] - i) % shape[0]) <= grid_range[0]))
        mask_j = ((np.abs((bin_indices[:, 1]) - j) <= grid_range[1]) |
                  ((-np.abs(bin_indices[:, 1] - j) % shape[1]) <= grid_range[1]))
        mask_k = ((np.abs((bin_indices[:, 2]) - k) <= grid_range[2]) |
                  ((-np.abs(bin_indices[:, 2] - k) % shape[2]) <= grid_range[2]))
        mask_ijk = (
            mask_i & mask_j & mask_k
        )
        return mask_ijk

    @staticmethod
    def _coordinates_to_bin_indices(coordinates: np.ndarray,
                                    range_x: List[float], width_x: float,
                                    range_y: List[float], width_y: float,
                                    range_z: List[float], width_z: float) -> np.ndarray:
        bin_indices = np.full((len(coordinates), 3), -1, dtype=int)
        for i, (range_i, width) in enumerate(zip([range_x, range_y, range_z],
                                               [width_x, width_y, width_z])):
            mask = (coordinates[:, i] >= range_i[0]) & (coordinates[:, i] <= range_i[1])
            bin_indices[mask, i] = (coordinates[mask, i] - range_i[0]) // width
        return bin_indices

    def __del__(self):
        self.analysis_file.close()

    def _get_filaments_from_analysis_file_or_data_reader(self, save: bool) -> List[List[Filament]]:
        if 'filaments' in self.analysis_file:
            n_frames = len(self.get_time_in_steps())
            return Analyser._load_filaments_from_h5_group(self.analysis_file['filaments'],
                                                          n_frames)
        filaments = self.data_reader.get_filaments_all()
        if save:
            group = self.analysis_file.create_group('filaments')
            Analyser._write_filaments_to_h5_group(group, filaments)
        return filaments

    @staticmethod
    def _load_filaments_from_h5_group(group: h5py.Group, n_frames: int) -> List[List[Filament]]:
        filaments = []
        for f in range(n_frames):
            filaments_f = Analyser._load_filaments_from_h5_group_single_frame(group[str(f)])
            filaments.append(filaments_f)
        return filaments

    @staticmethod
    def _load_filaments_from_h5_group_single_frame(group: h5py.Group) -> List[Filament]:
        items = group['items'][:]
        motors = group['motors'][:]
        filament_tuples = Analyser._parse_items_and_motors_arrays_to_filament_tuples(items, motors)
        filaments = [Filament(list(ftuple[0]), list(ftuple[1])) for ftuple in filament_tuples]
        return filaments

    @staticmethod
    @numba.njit
    def _parse_items_and_motors_arrays_to_filament_tuples(
            items: np.ndarray, motors: np.ndarray
    ) -> _FilamentsT:
        tuples = numba.typed.List()
        assert len(items) == len(motors)
        _invalid = np.iinfo(items.dtype).max
        for i in range(len(items)):
            items_i = items[i]
            items_i = items_i[items_i != _invalid]
            motors_i = motors[i]
            motors_i = motors_i[motors_i != _invalid]
            tuples.append((items_i, motors_i))
        return tuples

    @staticmethod
    def _write_filaments_to_h5_group(group: h5py.Group, filaments: List[List[Filament]]):
        for frame in range(len(filaments)):
            filaments_t = filaments[frame]
            typed_list = numba.typed.List()
            [typed_list.append((np.array(f.items).astype('uint32'),
                                np.array(f.motors).astype('uint32'))) for f in filaments_t]
            group_t = group.create_group(str(frame))
            Analyser._write_filaments_to_h5_group_single_frame(group_t, typed_list)

    @staticmethod
    def _write_filaments_to_h5_group_single_frame(
            group: h5py.Group, filaments: _FilamentsT):
        max_idx, max_items, max_motors = Analyser._get_maximum_items_and_value_for_filaments(filaments)

        items, motors = Analyser._create_filament_item_and_motor_arrays(len(filaments),
                                                                        max_idx,
                                                                        max_items,
                                                                        max_motors)

        Analyser._fill_filament_item_and_motor_arrays(filaments, items, motors)

        group.create_dataset('items', data=items, compression='lzf')
        group.create_dataset('motors', data=motors, compression='lzf')

    @staticmethod
    @numba.njit
    def _fill_filament_item_and_motor_arrays(
            filaments: _FilamentsT,
            items: np.ndarray, motors: np.ndarray):
        for i, (items_i, motors_i) in enumerate(filaments):
            items[i, :len(items_i)] = items_i
            motors[i, :len(motors_i)] = motors_i

    @staticmethod
    @numba.njit
    def _get_maximum_items_and_value_for_filaments(
            filaments: _FilamentsT
    ) -> Tuple[int, int, int]:
        max_items = 0
        max_idx = 0
        max_motors = 0
        for items_i, motors_i in filaments:
            max_idx_f = max(items_i)
            if max_idx_f > max_idx:
                max_idx = max_idx_f
            if len(items_i) > max_items:
                max_items = len(items_i)
            if len(motors_i) > max_motors:
                max_motors = len(motors_i)
        return max_idx, max_items, max_motors

    @staticmethod
    def _create_filament_item_and_motor_arrays(
            n_filaments: int, max_idx: int,
            max_items: int, max_motors:int
    ) -> Tuple[np.ndarray, np.ndarray]:
        dtype = 'uint32'
        _invalid = np.iinfo(dtype).max
        if max_idx > _invalid:
            raise ValueError("can't handle indices larger than "
                             f"{_invalid}, which is the maximum "
                             "for 32 bit unsigned integers")

        items = np.full((n_filaments, max_items), _invalid, dtype=dtype)
        motors = np.full((n_filaments, max_motors), _invalid, dtype=dtype)
        return items, motors







from typing import Tuple, List, Any, Union
import os
import json

import h5py
import matplotlib.pyplot as plt
import numpy as np
from .._pipeline import Pipeline
from ...tools.experiment_configuration import ExperimentConfiguration
from ...tools.experiment_iterator import Simulation, GroupIterator


class BeadMSD(Pipeline):

    def __init__(self, experiment_configuration: ExperimentConfiguration,
                 skip: int,
                 overwrite: bool=False):
        super().__init__(experiment_configuration)
        self.skip = skip
        self.plot_files.update({
            'MSD': os.path.join(experiment_configuration['MSD'], 'ensemble_msds.svg'),
        })
        self.output_files.update({
            'MSD': os.path.join(experiment_configuration['MSD'], 'msd_ensemble.h5'),
        })

        if overwrite:
            self._remove_output_files()

    def run_analysis(self):
        msd_data = self.get_ensemble_msds()
        self._plot_msds(msd_data)

    def _validate_configuration(self):
        assert "MSD" in self.experiment_configuration

    def get_ensemble_msds(self) -> 'EnsembleMSDData':
        global_max = 0
        iterator = self.experiment_configuration.experiment_iterator
        data = EnsembleMSDData()
        for group in iterator:
            selected_simulations = [sim for sim in group]
            lag_time, msd_ensemble = self._get_ensemble_msds_group(group, selected_simulations)
            global_max = max(global_max, msd_ensemble.max())

            radius, bead_diffusion_const = self._load_bead_diffusion_const(selected_simulations[0])
            data.append_data(lag_time, msd_ensemble, radius, bead_diffusion_const, group.label, group.color)
        return data

    def _plot_msds(self, data: 'EnsembleMSDData'):
        filename = self.plot_files['MSD']
        data.plot(filename)

    def _load_bead_diffusion_const(self, sim: Simulation) -> Tuple[float, float]:
        init_idx = sim.data_reader.read_parameter('initial_state_index')
        init_config = os.path.join(self.experiment_configuration.root_folder,
                                   f'initial_states/network{init_idx:02}/config.json')
        with open(init_config, 'rt') as fh:
            radius = json.load(fh)['acceptance_rate_handler']['bead_radius']
            bead_diffusion_const = 0.5 / radius
        return radius, bead_diffusion_const

    def _get_ensemble_msds_group(self, group: GroupIterator,
                                 selected_simulations) -> Tuple[np.ndarray, np.ndarray]:
        h5_group_template = group.get_label_from_values() + f"_skip_{self.skip}"
        h5_group_name = h5_group_template.format(**group.values)
        filename = self.output_files['MSD']
        lag_time, msd_ensemble = BeadMSD._load_ensemble_msds(filename, h5_group_name)
        if lag_time is not None:
            return lag_time, msd_ensemble

        lag_time, msds = self._get_individual_msds(selected_simulations)
        msd_ensemble = np.nanmean(msds, axis=0)
        BeadMSD._write_ensemble_msds(filename, h5_group_name, lag_time, msd_ensemble)
        return lag_time, msd_ensemble

    @staticmethod
    def _write_ensemble_msds(filename, h5_group_name, lag_time, msd_ensemble):
        with h5py.File(filename) as msds_file:
            h5_group = msds_file.create_group(h5_group_name)
            h5_group['log_msd'] = msd_ensemble
            h5_group['log_lag_time'] = lag_time

    @staticmethod
    def _load_ensemble_msds(
            filename: str,
            h5_group_name: str
    ) -> Union[Tuple[np.ndarray, np.ndarray], Tuple[None, None]]:
        if not os.path.exists(filename):
            return None, None
        with h5py.File(filename, 'r') as msds_file:
            if h5_group_name not in msds_file:
                return None, None
            h5_group = msds_file[h5_group_name]
            if 'log_msd' not in h5_group:
                return None, None
            return h5_group['log_lag_time'][:], h5_group['log_msd'][:]


    def _get_individual_msds(
            self, simulations: List[Simulation]
    ) -> Tuple[np.ndarray, np.ndarray]:
        lag_time = None
        msds = []
        for sim in simulations:
            dr = sim.data_reader
            a = sim.analyser
            msd_i = a.get_msd([0], non_filament=True, skip=self.skip)

            lags = msd_i[0, :, 0]
            msd_i = msd_i[0, :, 1]

            t_frames_steps = a.get_time_in_steps()
            dt = dr.read_dt() * (t_frames_steps[1] - t_frames_steps[0])
            if lag_time is None:
                lag_time = lags * dt
            else:
                assert (lag_time == lags * dt).all()
            msds.append(msd_i)
        return lag_time, np.vstack(msds)


class EnsembleMSDData:

    def __init__(self):
        self.lag_times = []
        self.msds = []
        self.bead_radii = []
        self.bead_diffusion_constants = []
        self.labels = []
        self.colors = []

    def append_data(self, lag_time, msd, bead_radius, bead_diffusion_constant, label, color):
        self.lag_times.append(lag_time)
        self.msds.append(msd)
        self.bead_radii.append(bead_radius)
        self.bead_diffusion_constants.append(bead_diffusion_constant)
        self.labels.append(label)
        self.colors.append(color)

    def get_lag_time_and_msd(self, index: int) -> Tuple[np.ndarray, np.ndarray]:
        return self.lag_times[index], self.msds[index]

    @property
    def length(self) -> int:
        return len(self.lag_times)

    def plot(self, filename: str) -> Tuple[plt.Figure, plt.Axes]:
        fig, ax = plt.subplots(1, 1, figsize=(11, 6))

        global_max = 0
        for i in range(self.length):
            msd = self.msds[i]
            global_max = max(global_max, msd.max())
            ax.plot(
                self.lag_times[i],
                msd,
                color=self.colors[i],
                label=self.labels[i]
            )
        colors, unique_free_diffusions = self._get_unique_free_diffusions()
        for i, (c, u) in enumerate(zip(colors, unique_free_diffusions)):
            ax.plot(self.lag_times[i], u, '--', color=c,
                    label='free diff.')

        ax.set(
            xscale='log',
            yscale='log',
            xlabel='$\\Delta t / t_0$',
            ylabel='MSD $\\langle \\Delta r^2 \\rangle / x_0^2$',
        )
        ax.legend(bbox_to_anchor=[1.05, 1.0], loc='upper left')
        ylim = ax.get_ylim()
        ax.set_ylim(ylim[0], global_max * 1.2)
        fig.tight_layout()
        fig.savefig(filename)
        return fig, ax

    def _get_unique_free_diffusions(self) -> Tuple[List[Any], List[np.ndarray]]:
        diff_consts = np.array(self.bead_diffusion_constants)
        uniques = np.unique(diff_consts)
        if len(uniques) == 1:
            return ['black'], [6 * self.lag_times[0] * uniques[0]]
        elif len(uniques) == self.length:
            return (self.colors,
                    [6 * self.lag_times[i] * self.bead_diffusion_constants[i] for i in range(self.length)])
        raise RuntimeError("Can only handle cases where all beads have same size or all beads have different size."
                           f" Your experiment has {self.length} ensemble MSDs "
                           f"and {len(uniques)} different bead sizes.")





class Box:

    def __init__(self):
        self.box_sizes = None

    def get_parameters_from_config_reader(self, config_reader):
        self.box_sizes = config_reader.get_box_sizes()

    def get_distance_vector_periodic_boundary(self, r1, r2):
        """
        r1 and r2 need to be numpy.ndarrays of shape (3,).
        """
        d = r2 - r1
        for i in range(3):
            d_i = d[i]
            size_i = self.box_sizes[i]
            half_size = size_i / 2
            if d_i > half_size:
                d[i] = d[i] - size_i
            elif d_i < -half_size:
                d[i] = d[i] + size_i
        return d



from typing import List, Tuple, Union, Dict, Any, Type, Generator
from dataclasses import dataclass
import os
import pandas as pd
import numpy as np
from matplotlib.colors import Colormap
from actomyosin_analyser.file_io.data_reader import DataReader
from ..analysis.analyser import Analyser


class ExperimentIterator:

    def __init__(self, data_reader_class: Type[DataReader],
                 group_by: List[Tuple[pd.DataFrame, List[str]]],
                 simulation_folder_template: str, exclude: List[int]):

        index0 = group_by[0][0].index
        previous_groups = [GroupIterator(data_reader_class, index0, {}, simulation_folder_template)]
        for table, columns in group_by:
            assert (table.index == index0).all()
            for col in columns:
                _groups = []
                for group in previous_groups:
                    sim_indices = np.array([idx for idx in group.index if idx not in exclude])
                    if len(sim_indices) == 0:
                        continue
                    selected = table.loc[sim_indices]
                    for value in selected[col].unique():
                        new_values = group.values.copy()
                        new_values.update({col: value})
                        new_group = GroupIterator(
                            data_reader_class=data_reader_class,
                            index=selected.index[selected[col] == value],
                            values=new_values,
                            simulation_folder_template=simulation_folder_template
                        )
                        _groups.append(new_group)
                previous_groups = _groups
        self.groups = previous_groups

    def __iter__(self):
        return iter(self.groups)

    def __len__(self):
        return len(self.groups)

    def get(self, **kwargs) -> List['GroupIterator']:
        selected = []
        for g in self.groups:
            if not g.has_matching_values(**kwargs):
                continue
            selected.append(g)
        return selected

    def create_index_from_groups(self) -> pd.Index:
        values = self.groups[0].values
        names = list(values.keys())
        arrays = []
        for n in names:
            arrays.append([g.values[n] for g in self.groups])
        if len(names) == 1:
            return pd.Index(arrays[0], name=names[0])
        return pd.MultiIndex.from_arrays(arrays, names=names)

    def assign_group_colors(
            self,
            key: str,
            colormap: Colormap,
            by_order: bool=True
    ):
        if by_order:
            self._assign_group_colors_by_order(key, colormap)
        else:
            self._assign_group_colors_by_value(key, colormap)

    def assign_group_labels(
            self,
            template_string: str
    ):
        for g in self.groups:
            g.label = template_string.format(**g.values)

    def _assign_group_colors_by_order(self, key, colormap):
        sorted_values = sorted([g.values[key] for g in self.groups])
        for g in self.groups:
            g.color = colormap(sorted_values.index(g.values[key]))

    def _assign_group_colors_by_value(self, key, colormap):
        for g in self.groups:
            g.color = colormap(g.values[key])


class GroupIterator:

    def __init__(self,
                 data_reader_class: Type[DataReader],
                 index: Union[pd.Index, pd.MultiIndex],
                 values: Dict[str, Any],
                 simulation_folder_template: str):
        self.data_reader_class = data_reader_class
        self.index = index
        self.values = values
        self._simulation_folder_template = simulation_folder_template
        self.color = None
        self.label = None

    def __iter__(self):
        return self._generator()

    def __len__(self):
        return len(self.index)

    def _generator(self) -> Generator['Simulation', None, None]:
        for idx in self.index:
            path = self._simulation_folder_template.format(idx)
            dr = self.data_reader_class(os.path.join(path, 'data.h5'))
            yield Simulation(
                index=idx,
                path=path,
                data_reader=dr,
                analyser=Analyser(dr, os.path.join(path, 'analysis.h5'))
            )

    def has_matching_values(self, **kwargs):
        for key in kwargs:
            if key not in self.values:
                raise KeyError(f'Groups are not created with key {key}')
            if self.values[key] != kwargs[key]:
                return False
        return True

    def __repr__(self):
        return (f'GroupIterator: Index = {self.index}; '
                f'selected by values {self.values}')

    def __str__(self):
        return self.__repr__()

    def get_label_from_values(self) -> str:
        label = ''
        for key in self.values:
            v = self.values[key]
            label = label + key + '_' + str(v) + '_'
        return '_'.join(label[:-1].split())


@dataclass
class Simulation:
    index: int
    path: str
    data_reader: DataReader
    analyser: Analyser



import numpy as np
from typing import List
from .bead import (Bead, Filament,
                   STATE_FILAMENT_START,
                   STATE_FILAMENT_INTERIOR,
                   STATE_FILAMENT_END)


class BeadStates(object):

    def __init__(self, links_array):
        self.linked_beads = {}
        self.filaments = []
        self.links = links_array
        
        self.states = self._bead_links_to_states(self.links)
        self.parse_links_to_beads(links_array)
        self.parse_beads_to_filaments()

    def get_indices_of_filament(self, filament_index: int) -> List[int]:
        f = self.filaments[filament_index]
        return f.items

    def parse_links_to_beads(self, links_array):
        if (links_array.shape[0] % 3 != 0):
            msg = "links_array needs to be multiple of 3 " 
            msg += "(links to previous, next, and cross-linked beads)."            
            raise RuntimeError(msg)
        for i in range(links_array.shape[0] // 3):
            links_i = links_array[i * 3: (i + 1) * 3]            
            if (links_i != -1).any():
                if links_i[0] == links_i[1]:
                    msg = "errors in data: bead " + str(i) \
                          + " has same previous and next neighbour!"
                    raise RuntimeError(msg)
                b = Bead()
                b.previous, b.next, b.cross = links_i                
                self.linked_beads[i] = b

    def parse_beads_to_filaments(self):
        for i in self.linked_beads:
            b_i = self.linked_beads[i]
            if b_i.previous == -1 and b_i.next != -1:
                start = i
                n = b_i.next
                items = [start, n]
                motors = []
                b_next = self.linked_beads[n]
                while b_next.next != -1:
                    if b_next.cross != -1:
                        motors.append(n)
                    n = b_next.next
                    items.append(n)
                    b_next = self.linked_beads[n]
                f = Filament(items, motors)
                self.filaments.append(f)

    def _bead_links_to_states(self, links_array):
        if links_array.shape[0] % 3 != 0:
            msg = "links_array' length needs to be multiple of 3 "
            msg += "(links to previous, next beads and across filaments)."
            raise RuntimeError(msg)
        s = np.zeros(links_array.shape[0] // 3)
        for i in range(0, s.shape[0], 1):
            l = links_array[i * 3: (i + 1) * 3]
            if (l == -1).all():
                continue
            if (l[0] != -1) and (l[1] != -1):
                s[i] = STATE_FILAMENT_INTERIOR                
            elif (l[0] != -1) and (l[1] == -1):
                s[i] = STATE_FILAMENT_END
            elif (l[0] == -1) and (l[1] != -1):
                s[i] = STATE_FILAMENT_START
        return s



from dataclasses import dataclass
from typing import Union
import numpy as np
from scipy.signal import find_peaks


@dataclass
class G0Data:
    index: int
    omega: float
    G0: float


def compute_G0_via_saddle_point(omega: np.ndarray, G: np.ndarray) -> Union['G0Data', None]:
    location_saddle = _get_location_of_saddle_point(G.real)
    if len(location_saddle) != 1:
        print("No saddle or too many found.")
        print("Found:", len(location_saddle))
        return
    location_saddle = location_saddle[0]
    g0 = G0Data(location_saddle,
                (omega[location_saddle] + omega[location_saddle + 1]) / 2,
                (G.real[location_saddle] + G.real[location_saddle + 1]) / 2)
    return g0


def compute_G0_via_tangent_minimum(omega: np.ndarray, G: np.ndarray) -> 'G0Data':
    tangent = G.imag/G.real
    peaks, properties = find_peaks(-tangent)
    if len(peaks) != 1:
        raise RuntimeError('The tangent has no minimum')
    location = peaks[0]
    g0 = G0Data(location, omega[location], G.real[location])
    return g0


def _get_location_of_saddle_point(data: np.ndarray) -> np.ndarray:
    finite_diff = data[1:] - data[:-1]
    peaks, properties = find_peaks(finite_diff)
    return peaks




"""
USAGE: 
   o install in develop mode: navigate to the folder containing this file,
                              and type 'pip install --user -e .'.
                              (omit '--user' if you want to install for
                               all users)                           
"""
from setuptools import setup, find_packages

setup(name='actomyosin_analyser',
      version='1.9.1',
      description='Analysis tools for actomyosin data',
      url='',
      author='Ilyas Kuhlemann',
      author_email='ilyasp.ku@gmail.com',
      license='GPLv3',
      packages=find_packages('.'),
      entry_points={
          "console_scripts": [
          ],
          "gui_scripts": [
          ]
      },
      install_requires=['numpy',
                        "matplotlib",
                        "h5py",
                        "numba",
                        "pandas",
                        "scipy",
                        "importlib-metadata; python_version < '3.8'"],
      zip_safe=False)



import numpy as np

def minimum_image_vectors_to_filament_coords(v: np.ndarray):
    shape = list(v.shape)
    shape[0] = shape[0] + 1
    coords = np.zeros(shape)
    coords[1:] = np.cumsum(v, axis=0)
    return coords



from abc import ABC, abstractmethod
from typing import List, Tuple

import numpy as np
import matplotlib.pyplot as plt
import pandas as pd

from .._pipeline import Pipeline
from ...analysis.microrheology.plateau_modulus import G0Data, compute_G0_via_saddle_point
from ...file_io.tables import merge_and_save_table
from ...tools.experiment_configuration import ExperimentConfiguration


class _GPipeline(Pipeline, ABC):

    def __init__(
            self, experiment_configuration: ExperimentConfiguration,
            skip_frames: int
    ):
        super().__init__(experiment_configuration)
        self.skip = skip_frames

    @abstractmethod
    def get_omegas_and_Gs(self, *args, **kwargs):
        ...

    def _plot_Gs(
            self,
            omegas: List[np.ndarray],
            Gs: List[np.ndarray],
            save_file_name: str=None
    ) -> Tuple[plt.Figure, plt.Axes]:
        fig, ax = plt.subplots(1, 1)
        for i, g in enumerate(self.experiment_configuration.experiment_iterator):
            color = g.color
            label = g.label
            omega = omegas[i]
            G = Gs[i]
            ax.plot(omega, G.real, '-', color=color, label=label)
            ax.plot(omega, G.imag, '--', color=color)

        _GPipeline._decorate_G_plot(ax)
        if save_file_name is not None:
            ax.legend()
            fig.tight_layout()
            fig.savefig(save_file_name)
        return fig, ax

    def _plot_Gs_with_G0s(
            self,
            omegas: List[np.ndarray],
            Gs: List[np.ndarray],
            G0s: List[G0Data],
            save_file_name: str=None
    ) -> Tuple[plt.Figure, plt.Axes]:
        fig, ax = self._plot_Gs(omegas, Gs)
        _GPipeline._add_G0s_to_plot(ax, G0s, color='C1', label='$G_0$')
        if save_file_name is None:
            return fig, ax
        fig.legend('off')
        fig.legend()
        fig.tight_layout()
        fig.savefig(save_file_name)
        return fig, ax

    @staticmethod
    def _decorate_G_plot(ax: plt.Axes):
        ax.set(
            xlabel=r'$\omega / (1/t_0)$',
            ylabel='moduli $G / (\\mathrm{kT}/x_0^3))$',
            xscale='log',
            yscale='log',
        )

    @staticmethod
    def _add_G0s_to_plot(ax: plt.Axes, G0s: List[G0Data],
                         color: str, label: str):
        omegas = [G0.omega for G0 in G0s]
        _G0s = [G0.G0 for G0 in G0s]
        ax.scatter(omegas, _G0s, c=color, zorder=100, label=label)

    def _create_G0_results_table(self) -> pd.DataFrame:
        index = self._create_group_index_with_skip_index()
        df = pd.DataFrame(index=index, columns=["omega", "G0"])
        return df

    def _create_group_index_with_skip_index(self):
        group_index = self.experiment_configuration.experiment_iterator.create_index_from_groups()
        arrays = [group_index.get_level_values(i) for i in range(group_index.nlevels)]
        arrays.append(pd.Index([self.skip] * len(self.experiment_configuration.experiment_iterator),
                               name='frames skipped'))
        index = pd.MultiIndex.from_arrays(arrays)
        return index

    def _find_G0_via_saddle_point_method(
            self, omegas,
            Gs, save_file_name: str
    ) -> List[G0Data]:
        G0_results = self._create_G0_results_table()
        G0s = []
        for i in range(len(omegas)):
            g0 = compute_G0_via_saddle_point(omegas[i], Gs[i])
            G0s.append(g0)
            G0_results.iloc[i] = g0.omega, g0.G0
        merge_and_save_table(save_file_name, G0_results)
        return G0s



from dataclasses import dataclass
from typing import List, Tuple, Union
import os

import h5py
import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
from .filament_trajectories import FilamentTrajectories
from .._pipeline import Pipeline
from ...tools.experiment_configuration import ExperimentConfiguration
from ...tools.experiment_iterator import GroupIterator


class DisplacementDistribution(Pipeline):

    def run_analysis(self, fixed_filament_length: int, n_bins: int):
        histograms = self.get_displacement_histograms(fixed_filament_length, n_bins)
        gauss_fit_results = self.get_gauss_fits(fixed_filament_length, n_bins, histograms)
        self._plot_histograms(histograms, gauss_fit_results)

    def _validate_configuration(self):
        assert "FilamentDisplacement" in self.experiment_configuration

    def __init__(self, experiment_configuration: ExperimentConfiguration):
        super().__init__(experiment_configuration)
        self.output_files.update({
            "Histograms": os.path.join(self.experiment_configuration["FilamentDisplacement"],
                                       'displacement_histograms.h5')
        })
        self.plot_files.update({
            "Histogram": os.path.join(self.experiment_configuration["FilamentDisplacement"],
                                       'displacement_histogram_{label}.svg')
        })
        self.pipeline_trajectories = FilamentTrajectories(experiment_configuration)


    def get_displacement_histograms(
            self, fixed_filament_length: int, n_bins: int
    ) -> List['_HistogramData']:
        histograms = []
        for g in self.experiment_configuration.experiment_iterator:
            hg = self._get_histogram_group(g, fixed_filament_length, n_bins)
            histograms.append(hg)
        return histograms

    def _get_histogram_group(
            self, group: GroupIterator,
            fixed_filament_length: int,
            n_bins: int
    ) -> '_HistogramData':
        group_label = group.get_label_from_values()
        edges, histogram = self._load_histogram_group(group_label, fixed_filament_length, n_bins)
        if edges is not None:
            return _HistogramData(edges, histogram)
        time, trajectories = self.pipeline_trajectories.get_trajectories_of_single_group(
            group, fixed_filament_length, use_beads_of_filament=[0, fixed_filament_length-1])
        displacements = trajectories[:, 1:] - trajectories[:, :-1]
        histogram, edges = np.histogram(displacements.flatten(), bins=n_bins, density=True)

        self._save_histogram_group(group_label, fixed_filament_length, n_bins, edges, histogram)

        return _HistogramData(edges, histogram)

    def _load_histogram_group(
            self, label: str,
            fixed_filament_length: int,
            n_bins: int
    ) -> Union[Tuple[None, None], Tuple[np.ndarray, np.ndarray]]:
        filename = self.output_files["Histograms"]
        if not os.path.exists(filename):
            return None, None
        with h5py.File(filename, 'r') as h5_file:
            if label not in h5_file:
                return None, None
            group = h5_file[label]

            if group.attrs['fixed_filament_length'] != fixed_filament_length:
                raise ValueError(f"fixed_filament_length {group.attrs['fixed_filament_length']} stored in file is "
                                 f"different from your specified parameter ({fixed_filament_length}). A group has only one"
                                 " unambiguous filament length.")
            if str(n_bins) not in group:
                return None, None

            edges = group[str(n_bins)]['edges'][:]
            histogram = group[str(n_bins)]['histogram'][:]

        return edges, histogram

    def _save_histogram_group(self, label: str, fixed_filament_length: int,
                              n_bins: int, edges: np.ndarray, histogram: np.ndarray):
        filename = self.output_files["Histograms"]
        with h5py.File(filename, 'a') as h5_file:
            if label not in h5_file:
                group = h5_file.create_group(label)
                group.attrs['fixed_filament_length'] = fixed_filament_length
            else:
                group = h5_file[label]
                if group.attrs['fixed_filament_length'] != fixed_filament_length:
                    raise ValueError(f"fixed_filament_length {group.attrs['fixed_filament_length']} stored in file is "
                                     f"different from your specified parameter ({fixed_filament_length}). A group" 
                                     " has only one unambiguous filament length.")
            s_n_bins = str(n_bins)
            if s_n_bins in group:
                group_n = group[s_n_bins]
            else:
                group_n = group.create_group(s_n_bins)

            group_n['edges'] = edges
            group_n['histogram'] = histogram

    def _plot_histograms(self, histograms: List['_HistogramData'], gauss_fits: List['_GaussFitResult']):
        groups = self.experiment_configuration.experiment_iterator.groups
        for i, g in enumerate(groups):
            hist = histograms[i]
            gauss = gauss_fits[i]
            fig, ax = plt.subplots(1, 1)
            ax.step(hist.x, hist.histogram, where='mid')
            if gauss is not None:
                ax.plot(hist.x, gauss.predict(hist.x), label='gauss fit')
                ax.legend()
            ax.set(
                title=g.label,
                xlabel='displacements / $x_0$',
                ylabel='density'
            )
            fig.tight_layout()
            figname = self.plot_files['Histogram'].format(label=g.get_label_from_values())
            fig.savefig(figname)
            plt.close(fig)


    def get_gauss_fits(
            self, fixed_filament_length: int,
            n_bins: int, histograms: List['_HistogramData']=None
    ) -> List['_GaussFitResult']:
        if histograms is None:
            histograms = self.get_displacement_histograms(fixed_filament_length, n_bins)
        results = []
        for hist in histograms:
            r = _GaussFitResult.create_from_fit(hist.x, hist.histogram)
            results.append(r)
        return results



@dataclass
class _GaussFitResult:

    optimized_parameters: np.ndarray
    covariance_parameters: np.ndarray

    @staticmethod
    def _gauss(x: np.ndarray, mu: float, sigma: float, a: float) -> np.ndarray:
        return a*np.exp(-(x-mu)**2/2/sigma**2)

    def predict(self, x):
        return _GaussFitResult._gauss(x, *self.optimized_parameters)

    @staticmethod
    def create_from_fit(x: np.ndarray, y: np.ndarray) -> Union['_GaussFitResult', None]:
        try:
            popt, pcov = curve_fit(_GaussFitResult._gauss, x, y)
            return _GaussFitResult(popt, pcov)
        except RuntimeError:
            return None

    @property
    def error(self):
        return np.sqrt(np.diag(self.covariance_parameters))


@dataclass
class _HistogramData:
    edges: np.ndarray
    histogram: np.ndarray

    @property
    def x(self) -> np.ndarray:
        return (self.edges[1:] + self.edges[:-1]) / 2




from typing import Tuple, Union
import logging
import numpy as np
from scipy.optimize import leastsq
from scipy.special import gamma


def compute_G_advanced(lag_time: np.ndarray, msd: np.ndarray,
                       radius: float,
                       n_points_per_decade: int,
                       kT: float,
                       dimensionality: int, half_width: int) -> Tuple[np.ndarray, np.ndarray]:
    """
    Compute G for MSD with a linear-spaced log_lag_time.
    Based on the code of
    T. Maier, T. Haraszti (http://journal.chemistrycentral.com/content/6/1/144).
    """

    log_lag_time, log_msd = _get_logspaced_axes(lag_time, msd, n_points_per_decade)
    J = _compute_J(log_msd, radius, kT, dimensionality)

    weight = _get_gaussian_weight_window(half_width)
    len_w = len(weight)

    J, log_lag_time = _get_valid_values(J, log_lag_time)

    N = len(log_lag_time)
    omega = np.full(N, np.nan)
    alpha = np.full(N, np.nan)
    g = np.full(N, np.nan)

    for i in range(N):
        alpha_i, g_i, om_i = _compute_g_i(J, i, len_w, log_lag_time, weight, half_width)
        if g_i < 0 or not np.isfinite(g_i):
            print('invalid g_i for index i = {}'.format(i))
            continue
        g[i] = g_i
        omega[i] = om_i
        alpha[i] = alpha_i

    g2 = []
    omega2 = []

    for i in np.arange(half_width, N - half_width - 1):
        alphaI, alphaR, pre = _compute_alphaR_and_alphaI_and_pre(
            g, i, omega, weight, half_width)

        g_i = g[i] * pre * complex(alphaR, alphaI)

        g2.append(g_i)
        omega2.append(omega[i])

    return np.array(omega2), np.array(g2)


def compute_G_danny_seara(
        tau: np.ndarray,
        msd: np.ndarray,
        bead_radius: float,
        dimensionality: int,
        kT: float,
        width: float,
        n_points_per_decade: int,
        clip: float=0.03
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Re-implementation of the matlab code on Danny Seara's github page
    https://github.com/dsseara/microrheology/blob/master/src/Moduli/calc_G.m .
    Originally code there was written by John C. Crocker and U. Penn.
    """

    tau, msd = _get_logspaced_axes(tau, msd, n_points_per_decade)

    omega = 1/tau
    C = dimensionality*kT/(3*np.pi*bead_radius)
    pi_2_1 = np.pi/2 - 1

    m, d, dd = _logderive(tau, msd, width)

    G_laplace = C/((m * gamma(1+d)) * (1+(dd/2)))
    g, da, dda = _logderive(omega, G_laplace, width)

    alpha_real = np.cos((np.pi/2) * da) - pi_2_1 * da * dda
    alpha_imag = np.sin((np.pi/2) * da) - pi_2_1 * (1-da) * dda
    G_real = g * (1 / (1 + dda)) * alpha_real
    G_imag = g * (1 / (1 + dda)) * alpha_imag

    G_real[G_real < G_laplace * clip] = 0
    G_imag[G_imag < G_laplace * clip] = 0

    if np.abs(dd).max() > 0.15 or np.abs(dda).max() > 0.15:
        logging.warning("High curvature in data, moduli may be unreliable!")
    return omega, G_real + 1j * G_imag


def _logderive(
        x: np.ndarray,
        y: np.ndarray,
        width: float
):
    n_points = len(x)
    df = np.zeros(n_points)
    ddf = np.zeros(n_points)
    f2 = np.zeros(n_points)
    lx = np.log(x)
    ly = np.log(y)

    for i in range(n_points):
        w = np.exp(-(lx - lx[i])**2 / (2 * width**2))
        ww = w > 0.03
        p = np.polyfit(lx[ww], ly[ww], 2, w=w[ww])
        f2 = np.exp(p[2] + p[1] * lx[i] + p[0] * lx[i]**2)
        df[i] = p[1] + (2 * p[0] * lx[i])
        ddf[i] = 2 * p[0]

    return f2, df, ddf


def _compute_alphaR_and_alphaI_and_pre(g, i, omega, weight, width):
    _pi_2 = np.pi / 2
    _pi_2_1 = _pi_2 - 1
    fit = _fit_pow_weight(omega[i - width: i + width + 1], g[i - width: i + width + 1], weight)
    a, b, c = fit[0]
    beta = c
    alpha = b + 2 * c * np.log(omega[i])
    pre = 1.0 / (1 + beta)
    alphapi = _pi_2 * alpha
    alphaR = np.cos(alphapi - _pi_2_1 * alpha * beta)
    alphaI = np.sin(alphapi - _pi_2_1 * beta * (1 - alpha))
    return alphaI, alphaR, pre


def _compute_g_i(J, i, len_w, log_lag_time, weight, half_width):
    start = max(i - half_width, 0)
    end = min(i + half_width, len(J))
    start_kernel = max(len_w // 2 - i, 0)
    end_kernel = len_w // 2 + min(len(J) - i, len_w // 2)
    fit = _fit_pow_weight(
        log_lag_time[start: end],
        J[start: end],
        weight[start_kernel: end_kernel]
    )
    a, b, c = fit[0]
    fitted = fit[1]
    beta = 2 * c
    alpha_i = b + 2 * c * np.log(log_lag_time[i])
    om_i = 1 / log_lag_time[i]
    yi = fitted[i - start]
    g_i = 1 / (yi * gamma(1 + alpha_i))
    g_i = g_i / (1 + 0.5 * beta)
    return alpha_i, g_i, om_i


def _get_valid_values(J, log_lag_time):
    valid = (log_lag_time > 0) & np.isfinite(log_lag_time) & (J > 0) & np.isfinite(J)
    log_lag_time = log_lag_time[valid]
    J = J[valid]
    return J, log_lag_time


def _get_gaussian_weight_window(width):
    w = 1.0
    weight = np.log(np.arange(1, 2 * width + 2, dtype=float) / width) / w
    weight = np.exp(np.exp(-0.5 * weight ** 2))
    return weight


def _get_logspaced_axes(t, x, n_points_per_decade: Union[int, float]):
    if n_points_per_decade < 8:
        raise ValueError("number of points per decade has to be 8 or higher "
                         "for this method to work more reliably")
    ratio = t[-1]/t[0]
    powers = np.linspace(0, np.log10(ratio), int(n_points_per_decade * np.log10(ratio)) + 1)
    logspaced_t = 10**powers * t[0]
    logspaced_selection = np.searchsorted(t, logspaced_t)
    logspaced_selection = np.unique(logspaced_selection)
    logspaced_selection = logspaced_selection[logspaced_selection < len(t)]
    t = t[logspaced_selection]
    x = x[logspaced_selection]
    _validate_log_spacing_has_at_least_8_points_per_decade(t)
    return t, x


def _validate_log_spacing_has_at_least_8_points_per_decade(t: np.ndarray):
    """
    According to the code by Crocker and Penn that is on Danny Seara's
    github repository https://github.com/dsseara/microrheology/blob/master/src/Moduli/calc_G.m,
    we need at least 7-8 points per decade for this method to work.
    """
    ratio = t[-1]/t[0]
    powers = np.arange(0, int(np.log10(ratio)) + 1)
    decades = 10**powers
    edges = decades * t[0]
    for i in range(1, len(edges)):
        if ((t >= edges[i-1]) & (t < edges[i])).sum() < 8:
            raise RuntimeError("Need at least 8 data points per decade. The number of points"
                               " for your log-spacing is set too small.")


def _compute_J(msd, radius, kT, dim) -> np.ndarray:
    return 3/dim * np.pi * radius / kT * msd


def _fit_pow_weight(
        x: np.ndarray, y: np.ndarray, weight: np.ndarray
) -> Tuple[Tuple[float, float, float], np.ndarray, np.ndarray]:
    if len(x) != len(y) or len(x) != len(weight):
        msg = "arrays need to have same lengths"
        raise ValueError(msg)

    if len(x) < 3:
        msg = "need at least 3 data points"
        raise ValueError(msg)

    lx = np.log(x)
    ly = np.log(y)

    valid = np.isfinite(lx) & np.isfinite(ly)

    c0, b0, a0 = np.polyfit(lx[valid], ly[valid], 2)

    params = [np.exp(a0), b0, c0]

    fit = leastsq(
        _err_pow_weight,
        params,
        (x, y, weight),
        full_output=False,
        maxfev=int(1e6)
    )

    a, b, c = fit[0]

    fitted = a * np.exp(b * lx + c * lx ** 2)
    error = (fitted - y) ** 2

    return (a, b, c), fitted, error


def _err_pow_weight(parameters, x, y, weight):
    a, b, c = parameters
    lx = np.log(x)
    power = b * lx + c * lx**2
    if power.max() > 10 or power.max() * np.log(abs(a)) > 250:
        return np.full_like(x, 1e99)

    yy = a * np.exp(power)
    chi = (yy - y) * weight

    return chi
